/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.pc.swing;

import javax.swing.*;
import java.io.*;
import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * Use XLogStreamTextArea to write an input stream to the screen.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class XLogStreamTextArea extends JTextArea
{
    public interface ILogListener
    {
        void logEvent(XLogStreamTextArea source);
    }

    public interface ILogFilter
    {
        boolean blockLine(XLogStreamTextArea source, String line);
    }

    static private Timer sTimer = new Timer("Timer for XLogStreamTextArea", true);
    
    private class XUpdateTask extends TimerTask
    {
        final static String NEWLINE ="\n";
        
        // Instanzvariablen
        BufferedInputStream mBIS;
        InputStreamReader mISR;
        BufferedReader mBR;
        AtomicBoolean mBlocked = new AtomicBoolean(false);
 
        XUpdateTask(InputStream is)
        {
            super();
            mBIS = new BufferedInputStream(is); // verwende für den InputStream einen Puffer
            mISR = new InputStreamReader(mBIS); // Reader, der den InputStream kapselt
            mBR = new BufferedReader(mISR);     // der Reader wird auch noch einmal gepuffert und bietet als
                                                // BufferedReader die Möglichkeit, eine ganze Zeile auszulesen.
        }
        
        boolean aquire()
        {
            return mBlocked.compareAndSet(false, true);
        }

        void release()
        {
            mBlocked.set(false);
        }
        
        public void run()
        {
            if (!aquire()) return;  // Stellt sicher, dass run() nicht parallel ausgeführt wird und weitere Threads nicht auf Ausführung warten.
            print();
            release();
        }

        void print()
        {
        	synchronized (mListeners)
        	{
                boolean newOutput = false;
	            try
	            {
	                String line = null;
	                while (mBR.ready() && (null != (line = mBR.readLine())))
	                {
	                	boolean blocked = (null == mFilter) ? false : mFilter.blockLine(XLogStreamTextArea.this, line);
	                	if (!blocked)
	                	{
		                    newOutput = true;
		                    append(line + NEWLINE);
	                	}
	                }
	            }
	            catch(IOException e) {}

	            if (newOutput)
	            {
	                for (ILogListener listener : mListeners)
	                {
	                    listener.logEvent(XLogStreamTextArea.this);
	                }
	            }
        	}
        }

        void finish()
        {
            while (!aquire())
            {
                try
                {
                    Thread.sleep(20);
                }
                catch(InterruptedException e) {}
            }
            print();          // final flush
            release();
            try
            {
                mBR.close();
                mISR.close();   // necessary?
                mBIS.close();   // necessary?
            }
            catch(IOException e) {}
        }
    }

    private XUpdateTask mUpdateTask = null;
    private ILogFilter  mFilter = null;
    private LinkedList<ILogListener> mListeners = new LinkedList<ILogListener>();

    public XLogStreamTextArea()
    {
        super();
    }

    public XLogStreamTextArea(String text)
    {
        super(text);
    }

    public XLogStreamTextArea(int rows, int columns)
    {
        super(rows, columns);
    }

    public XLogStreamTextArea(String text, int rows, int columns)
    {
        super(text, rows, columns);
    }

    public void startLogging(InputStream is)
    {
        stopLogging();
        mUpdateTask = new XUpdateTask(is);
        sTimer.schedule(mUpdateTask, 0, 40);
    }
    
    public void stopLogging()
    {
        if (null == mUpdateTask) return;
        
        mUpdateTask.cancel();   // mark task as completed
        sTimer.purge();         // remove task from timers task list
        mUpdateTask.finish();   // final cleanup of task
        mUpdateTask = null;
    }

    public void setFilter(ILogFilter filter)
    {
    	synchronized (mListeners)
    	{
    		mFilter = filter;
    	}
    }

    public void addListener(ILogListener listener)
    {
        synchronized (mListeners)
        {
            if (!mListeners.contains(listener)) mListeners.add(listener);
        }
    }

    public boolean removeListener(ILogListener listener)
    {
        synchronized (mListeners)
        {
            return mListeners.remove(listener);
        }
    }

    public void removeAllListeners()
    {
        synchronized (mListeners)
        {
            mListeners.clear();
        }
    }
}
