/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/

 
package org.jabotics.robot.core;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.adapter.XAdapter;

/**
 * CPilot is the internal controller class for motion control of the robot.
 * Higher level pilot classes are based on this class.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class CPilot
{
    // Klassenvariablen
    static private final IRobotAdapter sRobotAdapter = XAdapter.getFactory().getRobotAdapter();
    static private final float MIN_TRACK_WIDTH = 1f;
    static private final float MIN_WHEEL_DIAMETER = 0.4f;
    static private final float TRACK_WIDTH_DEFAULT = sRobotAdapter.getDefaultTrackWidth();
    static private final float WHEEL_DIAMETER_DEFAULT = sRobotAdapter.getDefaultWheelDiameter();
    static private final int MOTION_TYPE_POSITION = 0;
    static private final int MOTION_TYPE_SPEED = 1;
    static private final int MOTION_TYPE_STOP = 2;
    static private volatile CPilot sPilot = null;

    static private synchronized CPilot getPilot()
    {
        return (null == sPilot) ? getPilotSingleton(EMotorPort.B, EMotorPort.C) : sPilot; 
    }

    static public synchronized CPilot getPilotSingleton(EMotorPort portLeftMotor, EMotorPort portRightMotor)
    {
        if (null == sPilot)
        {
            sPilot = new CPilot(portLeftMotor, portRightMotor, WHEEL_DIAMETER_DEFAULT, TRACK_WIDTH_DEFAULT);
        }
        else
        {
            sPilot.assignMotors(portLeftMotor, portRightMotor, WHEEL_DIAMETER_DEFAULT, TRACK_WIDTH_DEFAULT);
        }

        return sPilot;
    }

    static public synchronized CPilot getPilotSingleton(EMotorPort portLeftMotor, EMotorPort portRightMotor, float wheelDiameter, float trackWidth)
    {
        if (null == sPilot)
        {
            sPilot = new CPilot(portLeftMotor, portRightMotor, wheelDiameter, trackWidth);
        }
        else
        {
            sPilot.assignMotors(portLeftMotor, portRightMotor, wheelDiameter, trackWidth);
        }

        return sPilot;
    }



    static public synchronized void initSession()
    {
        getPilot().initialize();
    }
    

    // Instanzvariablen
    private final Object mTachoMonitor;
    private final CMotor.XTransaction mMotorTransaction;
    private CMotor mLeftMotor;  
    private CMotor mRightMotor;
    private float mInitialWheelDiameter;  // in cm
    private float mInitialTrackWidth;     // in cm

    private float mWheelDiameter;       // in cm
    private float mWheelCircumference;  // in cm
    private float mTrackWidth;          // in cm
    private float mTurnCircumference;   // in cm
    private float mVTargetAbs;          // in °/s
    private float mATargetAbs;          // in °/s²
    private float mZeroPosLeft;         // in °
    private float mZeroPosRight;        // in °

    // For reinitialization of running motions on change of speed or acceleration
    private int mCurrentMotionType;
    private float mCurrentTargetLeft;
    private float mCurrentTargetRight;


    private CPilot(EMotorPort portLeftMotor, EMotorPort portRightMotor, float wheelDiameter, float trackWidth)
    {
        mTachoMonitor = new Object();
        assignMotors(portLeftMotor, portRightMotor, wheelDiameter, trackWidth);   // motors have to be initialized before adding this to session
        mMotorTransaction = new CMotor.XTransaction(mLeftMotor,mRightMotor);
    }

    synchronized private final void assignMotors(EMotorPort portLeftMotor, EMotorPort portRightMotor, float wheelDiameter, float trackWidth)
    {
        mLeftMotor = CMotor.getMotor(portLeftMotor);
        mRightMotor = CMotor.getMotor(portRightMotor);
        sRobotAdapter.setMotors(portLeftMotor,portRightMotor);
        mInitialWheelDiameter = wheelDiameter;
        mInitialTrackWidth = trackWidth;
        initialize();
    }

    synchronized private final void initialize()
    {
        // veränderliche Instanzvariablen (re)initialisieren
        setTrackWidth(mInitialTrackWidth);
        setWheelDiameter(mInitialWheelDiameter);
        mVTargetAbs = 0.5f * mLeftMotor.getMaxSpeed();
        mATargetAbs = 0.1f * mLeftMotor.getMaxAcceleration();
        resetCounters(); // initialization of mZeroPosLeft/Right
    }

    private void updateMotion()
    {
        if (!mMotorTransaction.motionCompleted())
        {
            switch (mCurrentMotionType)
            {
                case MOTION_TYPE_POSITION:
                    startMotionPosition(mCurrentTargetLeft,mCurrentTargetRight,false);
                    break;
                case MOTION_TYPE_SPEED:
                    float vLeftAbs  = Math.abs(mCurrentTargetLeft);
                    float vRightAbs = Math.abs(mCurrentTargetRight);
                    float factor = (vLeftAbs > vRightAbs) ? mVTargetAbs/vLeftAbs : mVTargetAbs/vRightAbs;
                    startMotionSpeed(factor*mCurrentTargetLeft,factor*mCurrentTargetRight);
                    break;
                default:
                    stopMotion(false);
            }
        }
    }

    private void startMotionPosition(float targetPosLeft, float targetPosRight, boolean waitUntilCompleted)
    {
        mCurrentMotionType = MOTION_TYPE_POSITION;
        mCurrentTargetLeft = targetPosLeft;
        mCurrentTargetRight = targetPosRight;

        float sLeft = targetPosLeft - mLeftMotor.getInternalPosition();     // in °
        float sRight = targetPosRight - mRightMotor.getInternalPosition();  // in °
        float sLeftAbs = Math.abs(sLeft);    // in °
        float sRightAbs = Math.abs(sRight);  // in °
        float signumLeft = (sLeft < 0) ? -1f : 1f;
        float signumRight = (sRight < 0) ? -1f : 1f;
        float weightLeftAbs = (sRightAbs > sLeftAbs) ? sLeftAbs/sRightAbs : 1f;
        float weightRightAbs = (sLeftAbs > sRightAbs) ? sRightAbs/sLeftAbs : 1f;
        float weightLeft = signumLeft * weightLeftAbs;
        float weightRight = signumRight * weightRightAbs;
        float vLeftAbs = weightLeftAbs * mVTargetAbs;
        float vRightAbs = weightRightAbs * mVTargetAbs;
        float aLeftAbs = weightLeftAbs * mATargetAbs;
        float aRightAbs = weightRightAbs * mATargetAbs;
        float vLeftIs = mLeftMotor.getSpeed();
        float vRightIs = mRightMotor.getSpeed();


        // Folgende Ansätze führen auf die Berechnungen von den Gewichtungsfaktoren entsprechenden Startgeschwindigkeiten
        // vLeftStart und vRightStart sowie von der Motorsteuerung zu korrigierender Fehlergeschwindigkeiten xVLiFehler und xVReFehler:
        //
        // Ansatz 1:   xVLiFehler == xVReFehler == xVFehler
        // (funktioniert insbesondere für signumLeft != signumRight; ObdA: weightLeftAbs == 1):
        //  (1a) vLeftIs = vLeftStart + xVFehler
        //  (1b) vRightIs = vRightStart + xVFehler
        //  (1c) vLeftStart * weightRight - vRightStart * weightLeft = 0   d.h. die Startgeschwindigkeiten stehen im gewünschten Verhältnis
        //  Durch Auflösen erhält man:
        //  (L1a) vLeftStart = (vLeftIs - vRightIs)/(1 - weightRight/weightLeft);  // ObdA: weightLeftAbs == 1
        //  (L1b) vRightStart = vLeftStart - (vLeftIs - vRightIs);
        //  (L1c) xVLiFehler = vLeftIs - vLeftStart;
        //  (L1d) xVReFehler = xVLiFehler    // = vRightIs - vRightStart;
        //
        // Ansatz 2:   xVLiFehler == -xVReFehler == xVFehler
        // (funktioniert insbesondere für signumLeft == signumRight; ObdA: weightLeftAbs == 1):
        //  (2a) vLeftIs = vLeftStart + xVFehler
        //  (2b) vRightIs = vRightStart - xVFehler
        //  (2c) vLeftStart * weightRight - vRightStart * weightLeft = 0   d.h. die Startgeschwindigkeiten stehen im gewünschten Verhältnis
        //  Durch Auflösen erhält man:
        //  (L2a) vLeftStart = (vLeftIs + vRightIs)/(1 + weightRight/weightLeft);  // ObdA: weightLeftAbs == 1
        //  (L2b) vRightStart = (vLeftIs + vRightIs) - vLeftStart;
        //  (L2c) xVLiFehler = vLeftIs - vLeftStart;
        //  (L2d) xVReFehler = xVLiFehler    // = vRightIs - vRightStart;
        //
        // Gewichtung der Ansätze:
        // Man verwendet eine Linearkombination der beiden Ansätze, um einen möglichst natürlichen Übergang
        // der Bewegungen zu erreichen. Da es praktisch auch auf eine schnelle Geschwindigkeitskorrektur ankommt,
        // werden die Lösungen der beiden Ansätze anhand ihrer Fehlergeschwindigkeiten gewichtet. Daraus ergibt
        // sich bei stetiger Veränderung der Ausgangssituation automatisch ein kontinuierliche Übergang des Roboterverhaltens.
        
        float vLeftStart = 0f;   // CMotor needs start velocity of regular motion.
        float vRightStart = 0f;  // CMotor calculates error velocity form intended start velocity and real velocity.
        float vSum = vLeftIs + vRightIs;
        float vDiff = vLeftIs - vRightIs;
        // Fallunterscheidungen mit der Brechstange:
        if (weightLeftAbs == weightRightAbs)
        {
            if (signumLeft == signumRight)
            {
                // special case: target motion is pure translation
                vLeftStart = 0.5f * vSum;
                vRightStart = vLeftStart;
                // => Korrektur eines bestehenden Rotationsanteils durch die Motorsteuerung
            }
            else
            {
                // special case: target motion is pure rotation
                vLeftStart = 0.5f * vDiff;
                vRightStart = -vLeftStart;
                // => Korrektur eines bestehenden Translationsanteils durch die Motorsteuerung
            }
        }
        else
        {
            // Für alle anderen Fälle sind prinzipiell beide Ansätze lösbar
            if (weightLeftAbs > weightRightAbs)
            {
                float relationRightLeft = weightRight/weightLeft;  // weightLeftAbs != 0
                // Ansatz 1:
                float vLeftStart1 = vDiff/(1 - relationRightLeft);  // relationRightLeft != 1
                float vRightStart1 = vLeftStart1 - vDiff;
                float errorV1 = (vLeftIs - vLeftStart);
                float errorV1Square = errorV1 * errorV1;
                // Ansatz 2:
                float vLeftStart2 = vSum/(1f + relationRightLeft);     // relationRightLeft != -1
                float vRightStart2 = vSum - vLeftStart2;
                float errorV2 = (vLeftIs - vLeftStart2);
                float errorV2Square = errorV2 * errorV2;
                // Gewichtung der Ansätze:
                float errorVSquareSum = errorV1Square + errorV2Square;
                if (errorVSquareSum == 0)
                {
                    vLeftStart = vLeftIs;
                    vRightStart = vRightIs;
                }
                else
                {
                    float weight1= errorV2Square / errorVSquareSum;
                    float weight2 = errorV1Square / errorVSquareSum;
                    vLeftStart = vLeftStart1 * weight1+ vLeftStart2 * weight2;
                    vRightStart = vRightStart1 * weight1+ vRightStart2 * weight2;
                }
            }
            else
            {
                // (weightLeftAbs < weightRightAbs)
                float relationLeftRight = weightLeft/weightRight;  // weightRightAbs != 0
                // Ansatz 1:
                float vRightStart1 = -vDiff/(1f - relationLeftRight);  // relationLeftRight != 1
                float vLeftStart1 = vRightStart1 + vDiff;
                float errorV1 = (vLeftIs - vLeftStart);
                float errorV1Square = errorV1 * errorV1;
                // Ansatz 2:
                float vRightStart2 = vSum/(1f + relationLeftRight);   // relationLeftRight != -1
                float vLeftStart2 = vSum - vRightStart2;
                float errorV2 = (vLeftIs - vLeftStart2);
                float errorV2Square = errorV2 * errorV2;
                // Gewichtung der Ansätze:
                float errorVSquareSum = errorV1Square + errorV2Square;
                if (errorVSquareSum == 0)
                {
                    vLeftStart = vLeftIs;
                    vRightStart = vRightIs;
                }
                else
                {
                    float weight1= errorV2Square / errorVSquareSum;
                    float weight2 = errorV1Square / errorVSquareSum;
                    vLeftStart = vLeftStart1 * weight1 + vLeftStart2 * weight2;
                    vRightStart = vRightStart1 * weight1 + vRightStart2 * weight2;
                }
            }
        }

        float vLeftCorrectionAbs = Math.abs(vLeftIs - vLeftStart);
        float vRightCorrectionAbs = Math.abs(vRightIs - vRightStart);
        float tStopCorrection = Math.max(0.001f,0.5f*Math.max(vLeftCorrectionAbs/mATargetAbs,vRightCorrectionAbs/mATargetAbs));

        mMotorTransaction.start();
        mLeftMotor.setTargetAcceleration(mMotorTransaction,aLeftAbs);
        mLeftMotor.setTargetSpeed(mMotorTransaction,vLeftAbs);
        mLeftMotor.startRotationPosLimited(mMotorTransaction,sLeft, vLeftStart, tStopCorrection);
        mRightMotor.setTargetAcceleration(mMotorTransaction,aRightAbs);
        mRightMotor.setTargetSpeed(mMotorTransaction,vRightAbs);
        mRightMotor.startRotationPosLimited(mMotorTransaction,sRight, vRightStart, tStopCorrection);
        mMotorTransaction.commit(waitUntilCompleted);
    }

    private void startMotionSpeed(float vLeftTarget, float vRightTarget)
    {
        mCurrentMotionType = MOTION_TYPE_SPEED;
        mCurrentTargetLeft = vLeftTarget;
        mCurrentTargetRight = vRightTarget;

        float vLeft = mLeftMotor.getSpeed();
        float vRight = mRightMotor.getSpeed();
        float dVLeftAbs = Math.abs(vLeftTarget - vLeft);
        float dVRightAbs = Math.abs(vRightTarget - vRight);

        mMotorTransaction.start();
        if ((0 != dVLeftAbs) || (0 != vRight))
        {
            float weightLeft = (dVRightAbs > dVLeftAbs) ? (dVLeftAbs/dVRightAbs) : 1;
            float weightRight = (dVLeftAbs > dVRightAbs) ? (dVRightAbs/dVLeftAbs) : 1;
            float aLeft = mATargetAbs*weightLeft;
            float aRight = mATargetAbs*weightRight;
            if (aLeft < CMotor.A_MIN) aLeft = CMotor.A_MIN;
            if (aRight < CMotor.A_MIN) aRight = CMotor.A_MIN;
            mLeftMotor.setTargetAcceleration(mMotorTransaction,aLeft);
            mRightMotor.setTargetAcceleration(mMotorTransaction,aRight);
        }
        else
        {
            mLeftMotor.setTargetAcceleration(mMotorTransaction,mATargetAbs);
            mRightMotor.setTargetAcceleration(mMotorTransaction,mATargetAbs);
        }
        if (vLeftTarget == 0)
        {
            mLeftMotor.stopRotation(mMotorTransaction);
        }
        else
        {
            mLeftMotor.setTargetSpeed(mMotorTransaction,vLeftTarget);
            mLeftMotor.startRotationUnlimited(mMotorTransaction,vLeftTarget >= 0);
        }
        if (vRightTarget == 0)
        {
            mRightMotor.stopRotation(mMotorTransaction);
        }
        else
        {
            mRightMotor.setTargetSpeed(mMotorTransaction,vRightTarget);
            mRightMotor.startRotationUnlimited(mMotorTransaction,vRightTarget >= 0);
        }
        mMotorTransaction.commit(false);
    }
    
    synchronized public void stopMotion(boolean waitUntilCompleted)
    {
        mCurrentMotionType = MOTION_TYPE_STOP;

        float vLeftAbs = Math.abs(mLeftMotor.getSpeed());
        float vRightAbs = Math.abs(mRightMotor.getSpeed());
        float weightLeft = (vRightAbs > vLeftAbs) ? (vLeftAbs/vRightAbs) : 1f;
        float weightRight = (vLeftAbs > vRightAbs) ? (vRightAbs/vLeftAbs) : 1f;
        float aLeft = mATargetAbs*weightLeft;
        float aRight = mATargetAbs*weightRight;
        if (aLeft < CMotor.A_MIN) aLeft = CMotor.A_MIN;
        if (aRight < CMotor.A_MIN) aRight = CMotor.A_MIN;

        mMotorTransaction.start();
        mLeftMotor.setTargetAcceleration(mMotorTransaction,aLeft);
        mRightMotor.setTargetAcceleration(mMotorTransaction,aRight);
        mLeftMotor.stopRotation(mMotorTransaction);
        mRightMotor.stopRotation(mMotorTransaction);
        mMotorTransaction.commit(waitUntilCompleted);
    }

    synchronized public void waitUntilCompleted()
    {
        mMotorTransaction.waitUntilMotionCompleted();
    }

    public float getWheelDiameter()
    {
        return mWheelDiameter;
    }

    synchronized public void setWheelDiameter(float diameter)
    {
        mWheelDiameter = (diameter < MIN_WHEEL_DIAMETER) ? MIN_WHEEL_DIAMETER : diameter;
        mWheelCircumference = D.PI_F * mWheelDiameter;

        sRobotAdapter.setWheelDiameter(mWheelDiameter);
    }

    public float getTrackWidth()
    {
        return mTrackWidth;
    }

    synchronized public void setTrackWidth(float trackWidth)
    {
        mTrackWidth = (trackWidth < MIN_TRACK_WIDTH) ? MIN_TRACK_WIDTH : trackWidth;
        mTurnCircumference = D.TWO_PI_F * mTrackWidth;

        sRobotAdapter.setTrackWidth(mTrackWidth);
    }
    
    public float getSpeed()
    {
        return (mWheelCircumference * (mRightMotor.getSpeed() + mLeftMotor.getSpeed()) / 720f);
    }

    public float getTargetSpeed()
    {
        return (mVTargetAbs * mWheelCircumference / 360f);
    }

    synchronized public void setTargetSpeed(float speed)
    {
        float vTarget = speed * 360f / mWheelCircumference;
        float vMax = 0.75f * mLeftMotor.getMaxSpeed();  // 25% reserved for gearing
        mVTargetAbs = (vTarget > vMax) ? vMax : ((vTarget < 20f) ? 20f : vTarget);
        updateMotion();
    }

    public float getMaxSpeed()
    {
        
        return (mWheelCircumference * mLeftMotor.getMaxSpeed() / 480f);  // 25% reserved for gearing => 0.75 * 1/360 = 1 / 480
    }

    synchronized public void setTargetSpeedToMax()
    {
        mVTargetAbs = 0.75f * mLeftMotor.getMaxSpeed();  // 25% reserved for gearing
        updateMotion();
    }

    public float getTargetAcceleration()
    {
        return (mATargetAbs * mWheelCircumference / 360f);
    }

    synchronized public void setTargetAcceleration(float acceleration)
    {
        float aTarget = acceleration * 360f / mWheelCircumference;
        float aMax = 0.75f * mLeftMotor.getMaxAcceleration();  // 25% reserved for gearing
        mATargetAbs = (aTarget > aMax) ? aMax : ((aTarget < 20f) ? 20f : aTarget);
        updateMotion();
    }

    public float getMaxAcceleration()
    {
        return (mWheelCircumference * mLeftMotor.getMaxAcceleration() / 480f);  // 25% reserved for gearing => 0.75 * 1/360 = 1 / 480
    }

    synchronized public void setTargetAccelerationToMax()
    {
        mATargetAbs = 0.75f * mLeftMotor.getMaxAcceleration();  // 25% reserved for gearing
        updateMotion();
    }

    public float getTachoCount()
    {
        return (mWheelCircumference * getTachoCountDegree() / 360f);
    }
    
    public float getOrientation()
    {
        return (getTachoDifferenceDegree() * mWheelCircumference / mTurnCircumference);
    }

    public void resetCounters()
    {
        synchronized (mTachoMonitor)
        {
            mZeroPosLeft = mLeftMotor.getInternalPosition();
            mZeroPosRight = mRightMotor.getInternalPosition();
        }
    }

    private float getTachoCountDegree()
    {
        synchronized (mTachoMonitor)
        {
            return (0.5f * (mLeftMotor.getInternalPosition() - mZeroPosLeft + mRightMotor.getInternalPosition() - mZeroPosRight));
        }
    }

    private float getTachoDifferenceDegree()
    {
        synchronized (mTachoMonitor)
        {
            return (mRightMotor.getInternalPosition() - mZeroPosRight - mLeftMotor.getInternalPosition() + mZeroPosLeft);
        }
    }

    synchronized public void moveStraight(boolean forward)
    {
        float v = forward ? mVTargetAbs : -mVTargetAbs;
        startMotionSpeed(v,v);
    }

    synchronized public void moveStraight(float distance, boolean waitUntilCompleted)
    {
        float dPos = 360f * distance / mWheelCircumference; 
        startMotionPosition(dPos + mLeftMotor.getInternalPosition(), dPos + mRightMotor.getInternalPosition(), waitUntilCompleted);
    }

    synchronized public void moveCurvature(float curvature, boolean forward)
    {
        if (100f < curvature)
        {
            curvature = 100f;
        }
        else if (-100f > curvature)
        {
            curvature = -100f;
        }
        
        float factorDS = 0.01f * curvature;
        float factorS = 1f - Math.abs(factorDS);
        float factorLeft = factorS - factorDS;
        float factorRight = factorS + factorDS;
        float weight = Math.max(Math.abs(factorLeft),Math.abs(factorRight));
        startMotionSpeed(factorLeft*mVTargetAbs/weight, factorRight*mVTargetAbs/weight);
    }

    synchronized public void moveRadius(float radius, boolean forward)
    {
        if (0 == radius)
        {
            moveStraight(forward);
        }
        else
        {
            float radiusAbs = Math.abs(radius);
            float v = (forward) ? mVTargetAbs : -mVTargetAbs;
            float curvFactorAbs = 0.5f * mTrackWidth / radiusAbs;
            float innerFactor = (1f - curvFactorAbs) / (1f + curvFactorAbs);
            if (radius > 0)
            {
                startMotionSpeed(innerFactor*v,v);  // left curve
            }
            else
            {
                startMotionSpeed(v,innerFactor*v);  // right curve
            }
        }
    }

    synchronized public void moveArc(float distance, float rotationAngle, boolean waitUntilCompleted)
    {
        float dPosHalfDS = 0.5f * mTurnCircumference / mWheelCircumference * rotationAngle; // in °
        float dPos = 360f * distance / mWheelCircumference;
        startMotionPosition(dPos - dPosHalfDS + mLeftMotor.getInternalPosition(), dPos + dPosHalfDS + mRightMotor.getInternalPosition(), waitUntilCompleted);
    }

    public void rotateByAngle(float rotationAngle, boolean waitUntilCompleted)
    {
        moveArc(0f, rotationAngle, waitUntilCompleted);
    }

    public void rotateToOrientation(float orientation, boolean waitUntilCompleted)
    {
        moveArc(0f, orientation - getOrientation(), waitUntilCompleted);
    }

    public boolean isMoving()
    {
        return (mLeftMotor.isMoving() || mRightMotor.isMoving());
    }
}
