/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.de;


/**
 *  <span class="beginner">&bull;</span> Die Schnittstelle IMotor bietet Methoden zur Steuerung eines Motors.
 * Bei Methodenaufrufen sind die Einheiten der Parameter zu beachten:
 *  Drehwinkel und Winkelstände in ° (Grad),
 *  Geschwindigkeiten in °/s (Grad pro Sekunde),
 *  Beschleunigungen in °/s² (Grad pro Sekundequadrat) und
 *  Zeiten in ms (Millisekunden).
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */

public interface IMotor
{
    /**
     * <span class="advanced">&bull;</span> Setzt den Korrekturfaktor für den Winkelstand des Motors.
     * Ungenauigkeiten des Winkelsensors eines Motors führen zu Abweichungen der Roboterbewegung von der
     * Sollbewegung. Besonders auffällig und störend ist dies bei Geradeausfahrten des Roboters.
     * Statt einer geradlinigen Bewegung beobachtet man ggf. eine leichte Kurve. Mithilfe des
     * Korrekturfaktors lässt sich dieses Problem weitgehend beheben. Ein Korrekturfaktor kleiner als 1
     * führt zu einer Vergrößerung der vom Roboter ausgeführten Rotationswinkel und -geschwindigkeiten.
     * (Vorsicht: Maximalgeschwindigkeiten nicht überschreiten!) Ein Korrekturfaktor größer als 1
     * führt zu kleineren beobachteten Rotationswinkeln und -geschwindigkeiten.
     * Der Wert des Korrekturfaktors wird auf den Bereich zwischen 0,9 und 1,1 beschränkt.
     * 
     * @param   korrekturfaktor   Korrekturfaktor für den Winkelstand (Standardwert 1)
     */
    public void setzeKorrekturfaktorWinkelstand(float korrekturfaktor);

    /**
     * <span class="beginner">&bull;</span> Liefert die aktuelle Position des Motors als Winkel.
     * 
     * @return      der aktuelle Winkelstand des Motors in Grad
     */
    public float holeWinkelstand();
    
    /**
     * <span class="beginner">&bull;</span> Setzt den Winkelstandszähler auf 0.
     */
    public void setzeWinkelstandZurueck();
    
    /**
     * <span class="beginner">&bull;</span> Liefert die tatsächliche Momentanwinkelgeschwindigkeit.
     * Im Gegensatz zur Sollgeschwindigkeit, die immer positiv angegeben wird,
     * kann die Monentangeschwindigkeit (Ist-Geschwindigkeit) auch negativ sein.
     * Eine negative Momentangeschwindigkeit entspricht einer Rückwärtsdrehbewegung des Motors.
     * 
     * @return  die Momentanwinkelgeschwindigkeit in °/s
     */
    public float holeGeschwindigkeit();
    
    /**
     * <span class="beginner">&bull;</span> Liefert die eingestellte Soll-Winkelgeschwindigkeit.
     * Die Sollgeschwindigkeit wird immer positiv angegeben und macht deshalb keine Aussage
     * über die Bewegungsrichtung.
     * 
     * @return  die Soll-Winkelgeschwindigkeit in °/s
     */
    public float holeSollgeschwindigkeit();

    /**
     * <span class="beginner">&bull;</span> Setzt die Soll-Winkelgeschwindigkeit.
     * Die Sollgeschwindigkeit wird immer positiv angegeben und macht deshalb keine Aussage
     * über die Bewegungsrichtung. Wird die Sollgeschwindigkeit auf 0 gesetzt, wird die aktuelle
     * Bewegung endgültig beendet.
     * 
     * @param   winkelgeschwindigkeit   die Soll-Geschwindigkeit in °/s
     */
    public void setzeSollgeschwindigkeit(float winkelgeschwindigkeit);
    
    /**
     * <span class="beginner">&bull;</span> Liefert die eingestellte Soll-Winkelbeschleunigung.
     * Die Sollbeschleunigung wird immer positiv angegeben und macht deshalb keine Aussage
     * über die Bewegungsrichtung.
     * 
     * @return  die Soll-Beschleunigung in °/s²
     */
    public float holeSollbeschleunigung();

    /**
     * <span class="beginner">&bull;</span> Setzt die Soll-Winkelbeschleunigung.
     * Die Sollbeschleunigung wird immer positiv angegeben und macht deshalb keine Aussage
     * über die Bewegungsrichtung.
     * 
     * @param   winkelbeschleunigung   die Soll-Beschleunigung in °/s²
     */
    public void setzeSollbeschleunigung(float winkelbeschleunigung);

    /**
     * <span class="beginner">&bull;</span> Liefert die maximal mögliche Winkelgeschwindigkeit.
     * Die Maximalgeschwindigkeit ist von der Batteriespannung und damit vom Ladezustand des Akkus abhängig.
     * 
     * @return  die Maximalgeschwindigkeit in °/s
     */
    public float holeMaximalgeschwindigkeit();
    
    /**
     * <span class="beginner">&bull;</span> Setzt die Soll-Winkelgeschwindigkeit auf den maximal möglichen Wert.
     * Die Maximalgeschwindigkeit ist von der Batteriespannung und damit vom Ladezustand des Akkus abhängig.
     */
    public void setzeSollgeschwindigkeitMaximal();
    
    /**
     * <span class="beginner">&bull;</span> Liefert die maximal mögliche Winkelbeschleunigung.
     * 
     * @return  die Maximalbeschleunigung in °/s²
     */
    public float holeMaximalbeschleunigung();

    /**
     * <span class="beginner">&bull;</span> Setzt die Soll-Winkelbeschleunigung auf den maximal möglichen Wert.
     */
    public void setzeSollbeschleunigungMaximal();

    /**
     * <span class="beginner">&bull;</span> Startet eine zeitlich unbegrente Drehung des Motors.
     * Die Programmausführung wird sofort nach dem Methodenaufruf mit dem nächsten Programmbefehl fortgesetzt,
     * während der Motor parallel dazu die Drehbewegung ausführt.
     *
     * @param  richtung    Richtung der Drehbewegung (siehe {@link ERichtung})
     */
    public void drehe(ERichtung richtung);

    /**
     * <span class="beginner">&bull;</span> Startet eine zeitlich begrenzte Drehung des Motors.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Drehbewegung oder erst nach Ende der Drehbewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * Es sollten keine allzu hohen Erwartungen an die Genauigkeit der Zeitmessung gestellt werden.
     * Zum einen treten aus technischen Gründen vereinzelt Verzögerungen bei der Motorsteuerung auf.
     * Zum anderen kann die vorgegebene Dauer nicht eingehalten werden, wenn der Motor sich bei Aufruf der Methode
     * bereits in Bewegung befindet und die eingestellte Sollbeschleunigung zu klein ist, um den Motor
     * in der vorgegebenen Zeit auf die Geschwindigkeit 0°/s abzubremsen.
     *
     * @param  richtung    Richtung der Drehbewegung (siehe {@link ERichtung})
     * @param  dauer       Dauer der Motordrehung in Millisekunden
     * @param  wartemodus  legt fest, ob das Ende der Bewegung vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void dreheZeit(float dauer, ERichtung richtung, EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Drehung des Motors um einen vorgegebenen Drehwinkel.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Drehbewegung oder erst nach Ende der Drehbewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param  drehwinkel  Drehwinkel in °
     * @param  wartemodus  legt fest, ob das Ende der Bewegung vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void dreheUmWinkel(float drehwinkel, EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Drehung des Motors zum vorgegebenen Winkelstand.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Drehbewegung oder erst nach Ende der Drehbewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param  winkelstand  Zielwinkelstand in °
     * @param  wartemodus   legt fest, ob das Ende der Bewegung vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void dreheZuWinkelstand(float winkelstand, EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Wartet das Ende der aktuellen Drehbewegung des Motors ab.
     * Die Programmausführung wird erst dann mit dem nächsten Programmbefehl
     * fortgesetzt, wenn die aktuelle Drehbewegung abschlossen ist.
     */
    public void warteEndeAb();
    
    /**
     * <span class="beginner">&bull;</span> Veranlasst das sofortige Anhalten des Motors.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * des Abbremsvorgangs oder erst nach tatsächlichem Ende der laufenden Drehbewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Abbremsbewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param  wartemodus  legt fest, ob das Ende des Abbremsvorgangs vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void halteAn(EWartemodus wartemodus);
    
    /**
     * <span class="advanced">&bull;</span> Veranlasst die freie Drehbarkeit des Motors.
     * Der Motor kann dann von Außen bewegt werden, ohne dass die Motorsteuerung dagegen arbeitet.
     */
    public void gebeFrei();
    
    /**
     * <span class="beginner">&bull;</span> Prüft, ob der Motor gerade eine Bewegung ausführt.
     * 
     * @return  "true" falls in Bewegung, "false" sonst
     */
    public boolean istInBewegung();
    
    /**
     * <span class="advanced">&bull;</span> Prüft, ob die Motorsteuerung ausgesetzt ist.
     * Der Motor-Controller unterbricht die Steuerung, wenn die Abweichung des
     * Ist-Winkelstand vom berechneten Soll-Winkelstand für eine zu große Zeit
     * einen zu großen Wert aufweist.
     * 
     * @return  "true" falls Steuerung unterbrochen, "false" sonst
     */
    public boolean istAusgesetzt();
}
