/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.de;


/**
 * <span class="beginner">&bull;</span> Die Schnittstelle IPilot bietet Methoden zur Bewegungssteuerung des Roboters.
 * Ein Pilot-Objekt benötigt dazu den Zugriff auf den linken und den rechten Motor
 * sowie Informationen über den Raddurchmesser und den Radabstand
 * (vgl. {@link IObjektfabrik#erzeugePiloten(EMotoranschluss,EMotoranschluss,float,float)}).
 * Bei Methodenaufrufen ist zu beachten, dass sich die verwendeten Einheiten von
 * denen der Einzelmotorsteuerung mittels {@link IMotor} z.T. unterscheiden:
 *  Strecken und Tachostand in cm,
 *  Ausrichtungswinkel des Roboters und Ausrichtungsänderungen bei Kurvenfahrten in ° (Grad),
 *  Geschwindigkeiten in cm/s,
 *  Beschleunigungen in cm/s² und
 *  Zeiten in ms (Millisekunden).
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */

public interface IPilot
{
    /**
     * <span class="advanced">&bull;</span> Liefert den aktuell gespeicherten Raddurchmesser der Räder.
     * 
     * @return  Raddurchmesser in cm
     */
    public float holeRaddurchmesser();

    /**
     * <span class="advanced">&bull;</span> Setzt den Raddurchmesser der Räder.
     * 
     * @param   durchmesser   der zu speichernde Raddurchmesser in cm
     */
    public void setzeRaddurchmesser(float durchmesser);

    /**
     * <span class="advanced">&bull;</span> Ersetzt durch die Methode {@link #holeSpurweite()}.
     * 
     * @return  Spurweite in cm
     */
    @Deprecated
    public float holeRadabstand();

    /**
     * <span class="advanced">&bull;</span> Ersetzt durch die Methode {@link #setzeSpurweite(float)}.
     * 
     * @param  spurweite  die zu speichernde Spurweite in cm
     */
    @Deprecated
    public void setzeRadabstand(float spurweite);

    /**
     * <span class="advanced">&bull;</span> Liefert die aktuell gespeicherte Spurweite der Antriebsräder.
     * 
     * @return  Spurweite in cm
     */
    public float holeSpurweite();

    /**
     * <span class="advanced">&bull;</span> Setzt die Spurweite der Antriebsräder.
     * 
     * @param  spurweite  die zu speichernde Spurweite in cm
     */
    public void setzeSpurweite(float spurweite);

    /**
     * <span class="beginner">&bull;</span> Liefert die tatsächliche Momentangeschwindigkeit des Roboterschwerpunkts.
     * Im Gegensatz zur Sollgeschwindigkeit, die immer positiv angegeben wird,
     * kann die Momentangeschwindigkeit (Ist-Geschwindigkeit) auch negativ sein.
     * Eine negative Momentangeschwindigkeit entspricht einer Rückwärtsbewegung des Roboters.
     * 
     * @return  die Momentangeschwindigkeit in cm/s
     */
    public float holeGeschwindigkeit();

    /**
     * <span class="beginner">&bull;</span> Liefert die eingestellte Sollgeschwindigkeit.
     * Die Sollgeschwindigkeit wird immer positiv angegeben und macht deshalb keine Aussage
     * über die Bewegungsrichtung.
     * 
     * @return  die Sollgeschwindigkeit in cm/s
     */
    public float holeSollgeschwindigkeit();

    /**
     * <span class="beginner">&bull;</span> Setzt die Sollgeschwindigkeit.
     * Die Sollgeschwindigkeit wird immer positiv angegeben und macht deshalb keine Aussage
     * über die Bewegungsrichtung. Es ist zu beachten, dass sich die Sollgeschwindigkeit
     * bei Kurvenfahrten auf das äußere Rad bezieht. Es gibt eine minimale Sollgeschwindigkeit
     * (unterhalb von 1 cm/s für den Standardroboter), die nicht unterschritten werden kann.
     * 
     * @param   geschwindigkeit   die Sollgeschwindigkeit in cm/s
     */
    public void setzeSollgeschwindigkeit(float geschwindigkeit);

    /**
     * <span class="beginner">&bull;</span> Liefert die maximal mögliche Geschwindigkeit.
     * Die Maximalgeschwindigkeit kann von der Batteriespannung und damit vom Ladezustand des Akkus abhängig sein.
     * 
     * @return  die Maximalgeschwindigkeit in cm/s
     */
    public float holeMaximalgeschwindigkeit();

    /**
     * <span class="beginner">&bull;</span> Setzt die Sollgeschwindigkeit auf den maximal möglichen Wert.
     * Die Maximalgeschwindigkeit kann von der Batteriespannung und damit vom Ladezustand des Akkus abhängig sein.
     */
    public void setzeSollgeschwindigkeitMaximal();

    /**
     * <span class="beginner">&bull;</span> Liefert die eingestellte Sollbeschleunigung.
     * Die Sollbeschleunigung wird immer positiv angegeben und macht deshalb keine Aussage
     * über die Bewegungsrichtung.
     * 
     * @return  die Soll-Beschleunigung in cm/s²
     */
    public float holeSollbeschleunigung();

    /**
     * <span class="beginner">&bull;</span> Setzt die Sollbeschleunigung.
     * Die Sollbeschleunigung wird immer positiv angegeben und macht deshalb keine Aussage
     * über die Bewegungsrichtung. Es gibt eine minimale Sollbeschleunigung
     * (wenige cm/s² für den Standardroboter), die nicht unterschritten werden kann.
     * 
     * @param   beschleunigung   die Sollbeschleunigung in cm/s²
     */
    public void setzeSollbeschleunigung(float beschleunigung);

    /**
     * <span class="beginner">&bull;</span> Liefert die maximal mögliche Beschleunigung.
     * 
     * @return  die Maximalbeschleunigung in cm/s²
     */
    public float holeMaximalbeschleunigung();

    /**
     * <span class="beginner">&bull;</span> Setzt die Sollbeschleunigung auf den maximal möglichen Wert.
     */
    public void setzeSollbeschleunigungMaximal();

    /**
     * <span class="beginner">&bull;</span> Liefert die aktuelle Position des Roboters.
     * 
     * @return      der aktuelle Tachostand in cm
     */
    public float holeTachostand();

    /**
     * <span class="beginner">&bull;</span> Liefert die aktuelle Ausrichtung des Roboters.
     * 
     * @return      der aktuelle Ausrichtungswinkel in °
     */
    public float holeAusrichtungswinkel();

    /**
     * <span class="beginner">&bull;</span> Setzt den Tachostand und den aktuellen Ausrichtungswinkel zurück.
     */
    public void setzeZaehlerstaendeZurueck();

    /**
     * <span class="beginner">&bull;</span> Startet eine zeitlich unbegrenzte geradlinige Fahrt mit vorgegebener Richtung.
     * Die Programmausführung wird sofort nach dem Start der Bewegung mit dem nächsten Programmbefehl fortgesetzt,
     * während der Roboter parallel dazu die Bewegung ausführt.
     *
     * @param  richtung    Richtung der Fahrt (siehe {@link ERichtung})
     */
    public void fahreGerade(ERichtung richtung);

    /**
     * <span class="beginner">&bull;</span> Startet eine geradlinige Bewegung zur Bewältigung einer vorgegebenen Strecke.
     * Eine negative Strecke entspricht einer Rückwärtsfahrt.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Bewegung oder erst nach Ende der Bewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param  strecke  die zurückzulegende Strecke in cm
     * @param  wartemodus  legt fest, ob das Ende der Bewegung vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void fahreGerade(float strecke, EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Startet eine zeitlich unbegrenzte Fahrt mit vorgegebenem Lenkungsparameter und vorgegebener Richtung.
     * Der Lenkungsparameter (ein Maß für die Krümmung) kann Werte zwischen -100 und 100 annehmen.
     * Positive Werte stehen für eine Linkskurve, negative für eine Rechtskurve.
     * Je größer der Bertrag des Lenkungsparameter, desto kleiner der Krümmungsradius
     * der Bahn. Ein Krümmungsparamter von 0 entspricht einer geradlinigen Bahn.
     * Ein Wert von 100 bzw. -100 entspricht einer Drehung auf der Stelle.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Bewegung oder erst nach Ende der Bewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param  lenkung   der Lenkungsparameter als Maß für die Stärke der Krümmung der Bahn
     * @param  richtung  Richtung der Fahrt (siehe {@link ERichtung})
     */
    public void fahreLenkung(float lenkung, ERichtung richtung);

    /**
     * <span class="beginner">&bull;</span> Startet eine zeitlich unbegrenzte Kurvenfahrt mit vorgegebenem Kurvenradius und vorgegebener Richtung.
     * Positive Werte des Kurvenradius stehen für eine Linkskurve (bezogen auf eine Vorwärtsfahrt), negative für eine Rechtskurve.
     * Die Programmausführung wird sofort nach dem Methodenaufruf mit dem nächsten Programmbefehl fortgesetzt,
     * während der Roboter parallel dazu die Bewegung ausführt.
     * 
     * @param  radius    der Radius der Kreisbahn in cm
     * @param  richtung  Richtung der Fahrt (siehe {@link ERichtung})
     */
    public void fahreRadius(float radius, ERichtung richtung);

    /**
     * <span class="beginner">&bull;</span> Startet eine Kurvenfahrt mit vorgegebenem Kurvenradius und vorgegebener Ausrichtungsänderung.
     * Beträgt der vorgegebene Radius 0 cm, so dreht sich der Roboter auf der Stelle um den angegebenen Winkel.
     * Aus der Radiusvorgabe und dem Änderungswinkel für die Roboterausrichtung wird eine entsprechende Bogenlänge
     * (d.h. Fahrtstrecke) berechnet. Beträgt der Winkel 0°, so entspricht dies einer Fahrtstrecke von 0 cm.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Bewegung oder erst nach Ende der Bewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param   radius      der Radius der Kreisbahn in cm
     * @param   drehwinkel  die Änderung des Ausrichtungswinkels in °
     * @param   wartemodus  legt fest, ob das Ende der Bewegung vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void fahreRadius(float radius, float drehwinkel, EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Startet eine Kurvenfahrt mit vorgegebener Strecke (Bogenlänge) und vorgegebener Ausrichtungsänderung.
     * Die angegebene Strecke bezieht sich auf den Roboterschwerpunkt.
     * Beträgt die vorgegebene Strecke 0 cm, so dreht sich der Roboter auf der Stelle
     * um den angegebenen Winkel. Beträgt der Winkel 0°, so handelt es sich um eine
     * geradlinige Bewegung. Andernfalls wird aus der Streckenvorgabe und dem
     * Änderungswinkel für die Roboterausrichtung ein entsprechender Kurvenradius berechnet.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Bewegung oder erst nach Ende der Bewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param   strecke     die zurückzulegende Strecke, d.h. Bogenlänge in cm
     * @param   drehwinkel  die Änderung des Ausrichtungswinkels in °
     * @param   wartemodus  legt fest, ob das Ende der Bewegung vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void fahreBogen(float strecke, float drehwinkel, EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Startet eine Drehung des Roboters auf der Stelle um den angegebenen Drehwinkel.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Bewegung oder erst nach Ende der Bewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param   drehwinkel  die Änderung des Ausrichtungswinkels in °
     * @param   wartemodus  legt fest, ob das Ende der Bewegung vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void dreheUmWinkel(float drehwinkel, EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Startet eine Drehung des Roboters auf der Stelle zum angegebenen Ausrichtungswinkel.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * der Bewegung oder erst nach Ende der Bewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Bewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param   ausrichtungswinkel  der Zielwinkel für die Ausrichtung des Roboters in °
     * @param   wartemodus          legt fest, ob das Ende der Bewegung vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void dreheZuAusrichtungswinkel(float ausrichtungswinkel, EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Wartet das Ende der aktuellen Bewegung ab.
     * Die Programmausführung wird also erst dann mit dem nächsten Programmbefehl
     * fortgesetzt, wenn die aktuelle Bewegung abschlossen ist.
     */
    public void warteEndeAb();

    /**
     * <span class="beginner">&bull;</span> Bricht die aktuelle Bewegung ab.
     * Über den Parameter wartemodus kann festgelegt werden, ob die Programmausführung sofort nach dem Starten
     * des Abbremsvorgangs oder erst nach tatsächlichem Ende der Bewegung mit dem nächsten Programmbefehl fortgesetzt wird.
     * Im ersten Fall werden die weiteren Programmbefehle also zeitlich parallel zur Abbremsbewegung ausgeführt.
     * Für weitere Erläuterungen zur parallelen Programmausführung siehe auch {@link EWartemodus}.
     * 
     * @param   wartemodus  legt fest, ob das Ende des Abbremsvorgangs vor der weiteren Programmausführung abgewartet wird oder nicht
     */
    public void halteAn(EWartemodus wartemodus);

    /**
     * <span class="beginner">&bull;</span> Prüft, ob der Roboter gerade eine Bewegung ausführt.
     * @since   1.0.5
     * 
     * @return  "true" falls in Bewegung, "false" sonst
     */
    public boolean istInBewegung();
}
