/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.compile.*;
import org.jabotics.robot.util.XLogBuffer;

/**
 * Beschreiben Sie hier die Klasse XMotor.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class XMotor extends BModelElement implements IMotorAdapter
{
    static double restrictedAngle(double angle)
    {
        return ((angle >= D.TWO_PI) || (angle < 0)) ? (angle - D.TWO_PI*Math.floor(angle/D.TWO_PI)) : angle;
    }

    static class RotationData
    {
        double getRotAngle(EMotorPort port)
        {
            return mRotAngle[port.ordinal()];
        }

        double getRotSpeed(EMotorPort port)
        {
            return mRotSpeed[port.ordinal()];
        }

        private double[] mRotAngle = new double[D.MOTOR_PORT_NUMBER]; // in rad
        private double[] mRotSpeed = new double[D.MOTOR_PORT_NUMBER]; // in rad / s
    }

    static class MotorSet
    {
    	private final XRobot mRobot;
        private final XMotor[] mMotors;
        private volatile boolean mExternalDataIsFrozen;
        private volatile RotationData mDataUpdate;        // temporary data used during dynamic update procedure (will replace sDataCurrent when done)
        private volatile RotationData mDataCurrent;       // currently valid data
        private volatile RotationData mDataExternal;      // data visible from outside (will reflect sDataCurrent, if not temporarily frozen for motor synchronization purposes)
        private volatile RotationData mDataGraphics;      // data version used for next graphics update
        private volatile double[] mPowerUpdate;
        private volatile double[] mPowerCommitted;
        private volatile double[] mPowerCurrent;

        MotorSet(XRobot robot)
        {
        	mRobot = robot;

        	initVolatileData();

            mMotors = new XMotor[D.MOTOR_PORT_NUMBER]; //D.MOTOR_PORT_NUMBER = EMotorPort.values().length;
            for (EMotorPort mp : EMotorPort.values())
            {
                mMotors[mp.ordinal()] = new XMotor(mRobot,mp,this); // Motors must be registered in this order at XSimulationModel for dynamics update
            }                                                       // => mMotors[0] will perform preparations for update of all motors
        }                                                           //    and mMotors[D.MOTOR_PORT_NUMBER] will terminate the update

        private void initVolatileData()
        {
            mExternalDataIsFrozen = false;
            mDataUpdate = new RotationData();
            mDataCurrent = mDataUpdate;
            mDataExternal = mDataUpdate;
            mDataGraphics = mDataUpdate;
            mPowerUpdate = new double[D.MOTOR_PORT_NUMBER];
            mPowerCommitted = mPowerUpdate;
            mPowerCurrent = mPowerUpdate;
        }

        public XMotor getMotor(EMotorPort port)
        {
            return mMotors[port.ordinal()];
        }

        public float[] getTachoCounts()
        {
	    	float[] tachoCounts = new float[D.MOTOR_PORT_NUMBER];
	    	for (EMotorPort port : EMotorPort.values()) tachoCounts[port.ordinal()] = mMotors[port.ordinal()].getTachoCount();
	    	return tachoCounts;
        }

        void prepareGraphicsUpdate(RotationData data)
        {
            mDataGraphics = data;
        }

        RotationData getCurrentData()
        {
            return mDataCurrent;
        }

        RotationData getGraphicsData()
        {
            return mDataGraphics;
        }

        private synchronized void updateRotationData()
        {
            if (!mExternalDataIsFrozen)
            {
                mDataExternal = mDataCurrent;
            }
        }

        synchronized void freezeRotationData()
        {
            if (!mExternalDataIsFrozen)
            {
                mExternalDataIsFrozen = true;
                mDataExternal = mDataCurrent;
            }
        }

        synchronized void releaseRotationData()
        {
            if (mExternalDataIsFrozen)
            {
                mExternalDataIsFrozen = false;
                mDataExternal = mDataCurrent;
            }
        }

        synchronized void commitPowerData()
        {
            double[] power = new double[D.MOTOR_PORT_NUMBER];
            for (int i = 0;  i < D.MOTOR_PORT_NUMBER;  ++i)  power[i] = mPowerUpdate[i];
            mPowerCommitted = power;
        }

        synchronized void updateDynamics()
        {
            mDataUpdate = new RotationData();
            mPowerCurrent = mPowerCommitted;
            for (EMotorPort port : EMotorPort.values())
            {
            	mMotors[port.ordinal()].updateDynamics();
            }
            mDataCurrent = mDataUpdate;
        }
    }



    // Instanzvariablen
    private XRobot mRobot;
    private final MotorSet mMotorSet;
    private final int mPortnumber;

    /**
     * Konstruktor für Objekte der Klasse XMotor
     */
    private XMotor(XRobot robot, EMotorPort port, MotorSet motor_set)
    {
    	// bare construction to be completed by setSession()
        super(robot.getSession());
        mRobot = robot;
        mMotorSet = motor_set;
        mPortnumber = port.ordinal();
    }
    
    @Override
    public synchronized void updateDynamics()
    {
        //*************************************************************************************************************
        // Aufzeichnungen von Beschleunigungs- und Auslaufbewegungen lieferten folgende empirische Ergebnisse:
        //*************************************************************************************************************
        // Das Ausrollen lässt sich für kleine Zeitschritte dt folgendermaßen beschreiben:
        //   v'=v*(1-15/s * dt)
        // bzw. noch genauer durch:
        //   v'=v*(1-14,5/s * dt *(10000 (°/s)² + v²) / (2000 (°/s)² + v²))
        // Der Einfluss des Leistungsparameters mPower lässt sich durch eine resultierende Beschleunigung erfassen:
        //   aPower = signum(mPower) * 13800°/s² * (mPower/100)²
        // Bei maximaler Leistung (mPower = 100) erreicht man eine gemessene Maximalgeschwindigkeit von mehr als 900°/s.
        // Die Simulation berechnet die neue Geschwindigkeit v' nach einem kleinen Zeitschritt dt also gemäß:
        //   v'=(v + aPower *dt) * (1-14,5/s * dt *(10000 (°/s)² + v²) / (2000 (°/s)² + v²))
        // Umrechnung der Zahlenwerte in rad, rad/s bzw. rad/s² liefert:
        //   aPower = 240.855 rad/s² * (mPower/100)² = 0,0240855 rad/s² * mPower²
        //   v'=(v + aPower *dt) * (1-14,5/s * dt *(3,0461742 (rad/s)² + v²) / (0,60923484 (°/s)² + v²))
        // Tests zeigen beim Ausrollen und zumindest für höhere Leistungen hervorragende Übereinstimmung mit
        // den Aufzeichnungen. Bei kleineren Leistungen funktioniert die Simulation der Hardware nicht
        // mehr ganz so perfekt, was sich aber in der Praxis durch den Regelungsmechanismus der Motoren
        // nicht bemerkbar machen sollte. Bereits bei einer Leistung von mPower = 60 (36% der physikalischen
        // Gesamtleistung) sind Unterschiede zwischen gemessenen und simulierten Daten klar erkennbar.
        //*************************************************************************************************************
        double power = mMotorSet.mPowerCurrent[mPortnumber];
        double a = 0.0240855 * power * power;
        if (power < 0) a = -a;
        double lastRotSpeed = mMotorSet.mDataCurrent.mRotSpeed[mPortnumber];
        double squareRotSpeed = lastRotSpeed * lastRotSpeed;
        double dt = mRobot.getDynamicUpdateTimePeriod();
        double rotSpeed = (lastRotSpeed + a * dt) * (1. - 14.5 * dt * (3.0461742 + squareRotSpeed) / (0.60923484 + squareRotSpeed));
        double meanRotSpeed = 0.5 * (rotSpeed + lastRotSpeed);
        mMotorSet.mDataUpdate.mRotAngle[mPortnumber] = mMotorSet.mDataCurrent.mRotAngle[mPortnumber] + meanRotSpeed * dt;
        mMotorSet.mDataUpdate.mRotSpeed[mPortnumber] = rotSpeed;
        
        if (CS.LOGGING) XLogBuffer.println(String.format("Updated:     SysTime: %d   Port: %d   Time: %d   Power: %f   Angle: %f   Speed: %f",mSession.currentTimeMillis(),mPortnumber,(int)(1000.*mRobot.getDynamicUpdateTime()),power,mMotorSet.mDataUpdate.mRotAngle[mPortnumber],rotSpeed));
    }

    @Override
    public void manageVisitor(BModelVisitor visitor)
    {
        visitor.visitMotor(this);
    }

    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // methods of IMotorAdapter
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public float getMaxSpeed()
    {
        return 800f;// Bei typischem Akku-Spannungswert von 8V Wert von:  lejos.nxt.Battery.getVoltage() * 100.0f;
    }

    @Override
    public void startControl()
    {
    }

    @Override
    public void setToFloatMode()
    {
        setPower(0);
    }

    @Override
    public void setPower(float power)
    {
    	mMotorSet.mPowerUpdate[mPortnumber] = power;
        if (CS.LOGGING) XLogBuffer.println(String.format("New Power:   SysTime: %d   Port: %d   Time: %d   Power: %f   Angle: %f   Speed: %f",mSession.currentTimeMillis(),mPortnumber,(int)(1000.*mRobot.getDynamicUpdateTime()),power,mMotorSet.mDataCurrent.mRotAngle[mPortnumber],mMotorSet.mDataCurrent.mRotSpeed[mPortnumber]));
    }

    @Override
    public float getTachoCount()
    {
    	mMotorSet.updateRotationData();
        return (float) (D.RAD_TO_DEGREE * mMotorSet.mDataExternal.mRotAngle[mPortnumber]);
    }

    @Override
    public float[] getPIDParameters()
    {
        return new float[] {6f,0.04f,22f,2f,0.04f,8f}; // P (control), I (control),D  (control), P (halted), I (halted), D (halted)
    }

    @Override
    public void setTachoCountCorrectionFactor(float correctionFactor)
    {
        // no correction necessary in "perfect" virtual world
    }

    @Override
    public void freezeAllTachoCounts()
    {
    	mMotorSet.freezeRotationData();
    }

    @Override
    public void releaseAllTachoCounts()
    {
    	mMotorSet.releaseRotationData();
    }

    @Override
    public void commitAllPowerData()
    {
    	mMotorSet.commitPowerData();
    }

    @Override
    public void initSession()
    {
        if (0 == mPortnumber)
        {
        	mMotorSet.initVolatileData();
        }
    }
}
