/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/

 
package org.jabotics.robot.sim;

import org.jabotics.pc.defs.SIMDefs;
import org.jabotics.robot.compile.CS;
import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.XLogBuffer;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.concurrent.atomic.AtomicLong;

import javax.swing.SwingUtilities;


/**
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class XSession implements Runnable
{
    static private final long MILLION = 1000000;
    static private final long CONFIRMATION_PERIOD = 3000000; // 3000000ns = 3ms

    
    static class XInitSessionVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.initSession();
        }
    }

    static class XStartSessionVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.startSession();
        }
    }

    static class XPauseSessionVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.pauseSession();
        }
    }


    static class XResumeSessionVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.resumeSession();
        }
    }

    static class XStopSessionVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.stopSession();
        }
    }

    static class XUpdateGraphicsVFC extends BModelVirtFuncCall
    {
        public void execute(BModelElement element)
        {
            element.updateGraphics();
        }
    }


    private class RobotThreadObserver implements Runnable
    {
        // Robot is master and running in a thread within this process

        @Override
        public void run()
        {
            try
            {
                mCallingThread.join();
            }
            catch (InterruptedException e) {}
            
            robotExecutionEnded();
        }
    }

    private class RobotThreadCreator implements Runnable
    {
        // Robot is NOT master and running in the owner thread of this Runnable instance:

        @Override
        public void run()
        {
            Class<?> robotClass;
            try
            {
                robotClass = Class.forName(mRobotClassName);
            }
            catch (ExceptionInInitializerError ie)
            {
                System.err.println(XMsg.INIT_CLASS_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(mRobotClassName);
                System.err.print(XMsg.ERROR_TYPE);
                System.err.println(" ExceptionInInitializerError");
                return;
            }
            catch (LinkageError le)
            {
                System.err.println(XMsg.INIT_CLASS_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(mRobotClassName);
                System.err.print(XMsg.ERROR_TYPE);
                System.err.println(" LinkageError");
                return;
            }
            catch (ClassNotFoundException cnfe)
            {
                System.err.println(XMsg.INIT_CLASS_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(mRobotClassName);
                System.err.print(XMsg.ERROR_TYPE);
                System.err.println(" ClassNotFoundException");
                return;
            }

            if (null == robotClass)
            {
                System.err.println(XMsg.INIT_CLASS_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(mRobotClassName);
                System.err.print(XMsg.ERROR_TYPE);
                System.err.println("Object construction failed.");
                return;
            }
            
            MethodHandles.Lookup lookup = MethodHandles.lookup();
            MethodType mt = MethodType.methodType(void.class, String[].class);
            MethodHandle mh;
            try
            {
                mh = lookup.findStatic(robotClass, "main", mt);
            }
            catch (NoSuchMethodException nsme)
            {
                System.err.println(XMsg.ACCESS_METHOD_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(mRobotClassName);
                System.err.print(XMsg.METHOD);
                System.err.println(" static main");
                System.err.print(XMsg.ERROR_TYPE);
                System.err.println(" NoSuchMethodException");
                return;
            }
            catch (IllegalAccessException iae)
            {
                System.err.println(XMsg.ACCESS_METHOD_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(mRobotClassName);
                System.err.print(XMsg.METHOD);
                System.err.println(" static main");
                System.err.print(XMsg.ERROR_TYPE);
                System.err.println(" IllegalAccessException");
                return;
            }
            if (null == mh)
            {
                System.err.println(XMsg.ACCESS_METHOD_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(mRobotClassName);
                System.err.print(XMsg.METHOD);
                System.err.println(" static main");
                System.err.print(XMsg.ERROR_TYPE);
                System.err.println(" Creation of method handle failed.");
                return;
            }

            if (CS.LOGGING_SIM) XLogBuffer.println("XSession.RobotThreadCreator.run(): Robot execution will be started now.");

            try
            {
                mh.invoke((Object[])null);
            }
            catch (Throwable t)
            {
                System.err.println(XMsg.ACCESS_METHOD_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(mRobotClassName);
                System.err.print(XMsg.METHOD);
                System.err.println(" static main");
                System.err.print(XMsg.ERROR_TYPE);
                System.err.print(" ");
                System.err.println(t.getMessage());
                return;
                //throw (t);
            }
            
            robotExecutionEnded();
        }
    }

    private class RobotProcessCreator implements Runnable
    {
        // Robot is running in separate process started and controlled by the simulation tool.

        @Override
        public void run()
        {
            // start robot process
            Process robotProcess = null;

            String java = System.getProperty("java.home") + (Env.WINDOWS ? "\\bin\\java.exe" : "/bin/java");
            ArrayList<String> commandList = new ArrayList<String>();
            ProcessBuilder pb = new ProcessBuilder(commandList);
            pb.directory(new File(mRobotRootPath));
            commandList.add(java);
            commandList.add("-D" + SIMDefs.KEY_SIM_UUID + "=" + Env.SIM_UUID);
            commandList.add("-D" + SIMDefs.KEY_SIM_SESSION_ID + "=" + mSessionID);
            commandList.add("-D" + SIMDefs.KEY_RMI_PORT + "=" + Env.RMI_PORT);
            commandList.add("-cp");
            commandList.add(Env.CLASS_PATH + File.pathSeparator + mRobotRootPath);
            commandList.add(mRobotClassName);
            try
            {
                robotProcess = pb.start();
            }
            catch (IOException e)
            {
                throw new InternalError("Internal error in XSession: Failed to start robot process.");
            }

            if (CS.LOGGING_SIM)
            {
                XLogBuffer.println("XSession.RobotProcessCreator.run(): Robot process has been started with parameters:");
                XLogBuffer.println("  java:        " + java);
                XLogBuffer.println("  uuid:        " + Env.SIM_UUID);
                XLogBuffer.println("  session id : " + mSessionID);
                XLogBuffer.println("  rmi port:    " + Env.RMI_PORT);
                XLogBuffer.println("  class name:  " + mRobotClassName);
                XLogBuffer.println("  root path:   " + mRobotRootPath);
            }

            // Redirect standard output and error stream of robot process to simulation tool.
            String line;
            BufferedInputStream bisOUT = new BufferedInputStream(robotProcess.getInputStream());
            InputStreamReader isrOUT = new InputStreamReader(bisOUT);
            BufferedReader brOUT = new BufferedReader(isrOUT);
            BufferedInputStream bisERR = new BufferedInputStream(robotProcess.getErrorStream());
            InputStreamReader isrERR = new InputStreamReader(bisERR);
            BufferedReader brERR = new BufferedReader(isrERR);

            boolean running = true;
            while (running)
            {
                msDelay(10);
                try
                {
                    while (brOUT.ready() && (null != (line = brOUT.readLine()))) System.out.println(line);
                }
                catch (IOException e) {}
                try
                {
                    while (brERR.ready() && (null != (line = brERR.readLine()))) System.err.println(line);
                }
                catch (IOException e) {}

                // check if process is still running
                try
                {
                    robotProcess.exitValue();
                    running = false;
                }
                catch (IllegalThreadStateException e) {}
            }
    
            robotExecutionEnded();
        }
    }

    /* SUSPENDING ROBOT THREAD
     * Idea:
     * Make use of deprecated methods Thread.suspend() and Thread.resume() for main robot thread.
     * Implementation of robot controller threads should make sure that no blocking issues occur
     * as long as implementation of robot class does not make unintended system calls.
     * Instead, provided class library should be used (e.g. via ITimerAdapter) and no additional
     * threads should be introduced.
     */
    private volatile ESessionState mState = ESessionState.NEW;
    private final Thread mCallingThread;
    private final int mSessionID;
    private final XRobot mRobot;
    private final String mRobotClassName;
    private final String mRobotRootPath;
    private final Object mRobotExecutionMonitor = new Object();
    private volatile boolean mRobotExecutionEnded;
    private volatile Thread mRobotMonitoringThread;
    private volatile Thread mSessionThread;
    private volatile boolean mIsPausing;
    private volatile long mTStart;
    private volatile long mTStartRobot;
    private volatile long mTNextConfirmation;
    private AtomicLong mTRobot = new AtomicLong();
    private TreeMap<String,ISessionHandler> mSessionHandlerMap = new TreeMap<String,ISessionHandler>();

    final XInitSessionVFC mInitSessionVFC = new XInitSessionVFC();
    final XStartSessionVFC mStartSessionVFC = new XStartSessionVFC();
    final XPauseSessionVFC mPauseSessionVFC = new XPauseSessionVFC();
    final XResumeSessionVFC mResumeSessionVFC = new XResumeSessionVFC();
    final XStopSessionVFC mStopSessionVFC = new XStopSessionVFC();
    final XUpdateGraphicsVFC mUpdateGraphicsVFC = new XUpdateGraphicsVFC();
    




    XSession(int session_id, String robotClassName, String robotRootPath)
    {
        mSessionID = session_id;
        mRobotClassName = robotClassName;
        mRobotRootPath = robotRootPath;
        mCallingThread = Thread.currentThread();
        mRobotMonitoringThread = null;
        mSessionThread = null;
        
        initVolatileData();

        XAudioSystem.setSession(this); // TODO: redesign to support multiple sessions
        mRobot = new XRobot(this);   // bare construction of robot and its child elements
        XSensorRef.setSession(this); // TODO: redesign to support multiple sessions

        XSimulationModel.addUpdateElement(mRobot);

        if (CS.LOGGING_SIM)
        {
            XLogBuffer.println("XSession(): Created with parameters:");
            XLogBuffer.println("  mSessionID:      " + mSessionID);
            XLogBuffer.println("  mRobotClassName: " + mRobotClassName);
            XLogBuffer.println("  mRobotRootPath:  " + mRobotRootPath);
        }
    }


    private void initVolatileData()
    {
        mRobotExecutionEnded = false;
        mIsPausing = true;
        mTStart = System.nanoTime();
        mTStartRobot = mTStart;
        mTNextConfirmation = mTStart + CONFIRMATION_PERIOD;
        mTRobot.set(0);
    }

    public int getID()
    {
        return mSessionID;
    }

    public XRobot getRobot()
    {
        return mRobot;
    }

    public ESessionState getState()
    {
        return mState;
    }


    synchronized void init()
    {
        if ((ESessionState.NEW != mState) && (ESessionState.COMPLETED != mState))
        {
            throw new IllegalStateException("XSession.init(): Current state is " + mState.toString() + " but has to be NEW or COMPLETED.");
        }

        initVolatileData();

        // traversal root to leaf because sensors rely on robot's dynamic data
        mRobot.traverseRootLeaf(mInitSessionVFC);

        Iterator<ISessionHandler> it = mSessionHandlerMap.values().iterator();
        while (it.hasNext()) it.next().initSession();

        mState = ESessionState.INITIALIZED;

        if (CS.LOGGING_SIM) XLogBuffer.println("XSession.init() completed.");
    }

    synchronized void start()
    {
        if (ESessionState.INITIALIZED != mState)
        {
            throw new IllegalStateException("XSession.start(): Current state is " + mState.toString() + " but has to be INITIALIZED.");
        }

        mTStartRobot = System.nanoTime();
        mIsPausing = false;

        mRobot.traverseRootLeaf(mStartSessionVFC);

        Iterator<ISessionHandler> it = mSessionHandlerMap.values().iterator();
        while (it.hasNext()) it.next().startSession();

        if (Env.ROBOT_IS_MASTER)
        {
            if (!Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
            {
                // Monitor master robot thread by the following thread:
                mRobotMonitoringThread = new Thread(new RobotThreadObserver());
                mRobotMonitoringThread.start();
            }

            // (Env.ROBOT_IS_MASTER && Env.ROBOT_RUNS_IN_SEPARATE_PROCESS):
            // SIMAdapter in robot process is responsible for sending notification
            // on end of robot execution (compare SIMAdapter.initialize()).
            // Therefore no monitoring thread here.
        }
        else
        {
            if (Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
            {
                // Start robot process monitored by the following thread:
                mRobotMonitoringThread = new Thread(new RobotProcessCreator());
                mRobotMonitoringThread.start();
            }
            else
            {
                // In this case the monitoring thread is the robot thread itself:
                mRobotMonitoringThread = new Thread(new RobotThreadCreator());
                mRobotMonitoringThread.start();
            }
        }

        if (null == mSessionThread)
        {
            // Final stop() will be triggered in mSessionThread by end of robot execution:
            mSessionThread = new Thread(this,"XSession.Session#" + mSessionID);
            mSessionThread.start();
        }

        mState = ESessionState.RUNNING;

        if (CS.LOGGING_SIM) XLogBuffer.println("XSession.start() completed.");
    }

    synchronized void pause()
    {
        if (ESessionState.RUNNING != mState)
        {
            throw new IllegalStateException("XSession.pause(): Current state is " + mState.toString() + " but has to be RUNNING.");
        }

        nanoTimeRobot(); // updates latest robot time in mTRobot
        mIsPausing = true;

        mRobot.traverseRootLeaf(mPauseSessionVFC);

        Iterator<ISessionHandler> it = mSessionHandlerMap.values().iterator();
        while (it.hasNext()) it.next().pauseSession();

        mState = ESessionState.PAUSING;

        if (CS.LOGGING_SIM) XLogBuffer.println("XSession.pause() completed.");
    }

    synchronized void resume()
    {
        if (ESessionState.PAUSING != mState)
        {
            throw new IllegalStateException("XSession.resume(): Current state is " + mState.toString() + " but has to be PAUSING.");
        }

        mTStartRobot = System.nanoTime() - mTRobot.get();
        mIsPausing = false;

        mRobot.traverseRootLeaf(mResumeSessionVFC);

        Iterator<ISessionHandler> it = mSessionHandlerMap.values().iterator();
        while (it.hasNext()) it.next().resumeSession();

        mState = ESessionState.RUNNING;

        if (CS.LOGGING_SIM) XLogBuffer.println("XSession.resume() completed.");
    }

    synchronized void stop()
    {
        if (ESessionState.RUNNING != mState)
        {
            throw new IllegalStateException("XSession.stop(): Current state is " + mState.toString() + " but has to be RUNNING.");
        }

        nanoTimeRobot(); // updates latest robot time in mTRobot
        mIsPausing = true;

        mRobot.traverseRootLeaf(mStopSessionVFC);

        Iterator<ISessionHandler> it = mSessionHandlerMap.values().iterator();
        while (it.hasNext()) it.next().stopSession();

        System.runFinalization(); // updates reference counting
        System.gc();

        mState = ESessionState.STOPPED;

        if (CS.LOGGING_SIM) XLogBuffer.println("XSession.stop() completed.");
    }

    synchronized void complete()
    {
        if (ESessionState.STOPPED != mState)
        {
            throw new IllegalStateException("XSession.cleanup(): Current state is " + mState.toString() + " but has to be STOPPED.");
        }

        // cleans up all session data and builds up new session handler map
        Iterator<ISessionHandler> it = mSessionHandlerMap.values().iterator();
        mSessionHandlerMap = new TreeMap<String,ISessionHandler>();
        while (it.hasNext())
        {
            ISessionHandler handler = it.next();
            it.remove();
            ISessionHandler newHandler = handler.completeSession();
            if (!handler.equals(newHandler)) handler.handlerRemoved(this);
            if (null != newHandler) this.addSessionHandler(newHandler);
        }

        if (!Env.ROBOT_IS_MASTER)
        {
            mSessionThread = null;
        }

        mState = ESessionState.COMPLETED;

        if (CS.LOGGING_SIM) XLogBuffer.println("XSession.cleanup() completed.");
    }

    synchronized void terminate()
    {
        mState = ESessionState.TERMINATING;

        Iterator<ISessionHandler> it = mSessionHandlerMap.values().iterator();
        while (it.hasNext()) it.next().terminateSession();

        if (null != mSessionThread)
        {
            try
            {
                mSessionThread.join(); // see implementation of run(): robot process will be stopped
            }
            catch (InterruptedException e) {}
        }

        if (CS.LOGGING_SIM) XLogBuffer.println("XSession.terminate() completed.");
    }

    void robotExecutionEnded()
    {
        synchronized(mRobotExecutionMonitor)
        {
            if (!mRobotExecutionEnded)
            {
                mRobotExecutionEnded = true;
                mRobotExecutionMonitor.notifyAll();
            }
        }
    }

    public boolean canBeRestarted()
    {
        return (null != mRobotClassName);
    }

    public synchronized void resetPreferences()
    {
        XPreferences.reset();
    }

    public synchronized void importPreferences(String pathname)
    {
        XPreferences.importFromXML(pathname);
    }

    public synchronized void importPreferencesByClass(String classname)
    {
        String pathname = classname.replace('.',java.io.File.separatorChar) + ".xml";
        File file = new File(pathname);
        if (file.isFile() && file.canRead())
        {
            XPreferences.importFromXML(file);
        }
        else
        {
            // working directory in Eclipse typically is project dir containing bin directory
            XPreferences.importFromXML("bin" + java.io.File.separatorChar + pathname);
        }
    }

    public void pressButton(EButtonType button_type, int period)
    {
        mRobot.getButton(button_type).setPressedForPeriod((period < 20) ? 20 : period);
    }

    // If a special initialization ordering is required during session handling,
    // this can be achieved by choosing appropriate keys for the different
    // session handlers.
    public void addSessionHandler(ISessionHandler handler)
    {
        if (null != handler)
        {
            String key = handler.getKey();
            synchronized(mSessionHandlerMap)
            {
                if (null != key)
                {
                    ISessionHandler replaced = mSessionHandlerMap.put(key,handler);
                    if (null == replaced)
                    {
                        handler.handlerAdded(this);
                    }
                    else if (!replaced.equals(handler))
                    {
                        replaced.handlerRemoved(this);
                        handler.handlerAdded(this);
                    }
                }
            }
        }
    }

    public ISessionHandler getSessionHandler(String key)
    {
        synchronized(mSessionHandlerMap)
        {
            return mSessionHandlerMap.get(key);
        }
    }

    public ISessionHandler removeSessionHandler(String key)
    {
        synchronized(mSessionHandlerMap)
        {
            ISessionHandler handler = mSessionHandlerMap.remove(key);
            if (null != handler) handler.handlerRemoved(this);
            return handler;
        }
    }

    public void clearSessionHandlers()
    {
        synchronized(mSessionHandlerMap)
        {
            Iterator<ISessionHandler> it = mSessionHandlerMap.values().iterator();
            while (it.hasNext())
            {
                ISessionHandler handler = it.next();
                it.remove();
                handler.handlerRemoved(this);
            }
        }
    }

    public XSessionData getSessionData()
    {
        return new XSessionData(mState, nanoTimeRobot(), XSimulationModel.getDynamicsUpdateCount());
    }

    public long currentTimeMillis()
    {
        return nanoTime() / MILLION;
    }

    public long nanoTime()
    {
        return System.nanoTime() - mTStart;
    }

    public static void msDelay(long period)
    {
        nsDelay(period * MILLION);
    }

    public static void nsDelay(long period)
    {
        if (period > 0)
        {
            int nanos = (int)(period % MILLION);
            long millis = (period - nanos) / MILLION;
            try
            {
                Thread.sleep(millis,nanos);
            }
            catch (InterruptedException e)
            {
                Thread.currentThread().interrupt();
            }
        }
    }

    public long currentTimeMillisRobot()
    {
        return nanoTimeRobot() / MILLION;
    }

    public long nanoTimeRobot()
    {
        if (mIsPausing) return mTRobot.get();

        long newRobotTime;
        long lastRobotTime;
        do
        {
            long nanoTime = System.nanoTime();
            long nextConfirmTime = mTNextConfirmation;
            newRobotTime = (nanoTime > nextConfirmTime) ? nextConfirmTime - mTStartRobot : nanoTime - mTStartRobot;
            lastRobotTime = mTRobot.get();
            if (newRobotTime <= lastRobotTime) return lastRobotTime;
        }
        while (!mTRobot.compareAndSet(lastRobotTime, newRobotTime));
        return newRobotTime;            
    }

    public void notifySimulationIsAlive()
    {
        long nanoTime = System.nanoTime();
        if (!mIsPausing)
        {
            long exceedingTime = nanoTime - mTNextConfirmation;
            if (exceedingTime > 0) mTStartRobot += exceedingTime;
        }
        mTNextConfirmation = nanoTime + CONFIRMATION_PERIOD;
    }

    public void msDelayRobot(long period)
    {
        nsDelayRobot(period * MILLION);
    }

    public void nsDelayRobot(long period)
    {
        long remaining = period;
        long start = nanoTimeRobot();
        while (remaining > 0)
        {
            nsDelay(remaining);
            if (Thread.currentThread().isInterrupted()) return;
            remaining = period - (nanoTimeRobot() - start);
        }
    }


    public void updateGraphics(XRobot.XDynamicData dd)
    {
        mRobot.prepareGraphicsUpdate(dd);
        mRobot.traverseLeafRoot(mUpdateGraphicsVFC);
    }


    /****************************************************************************************
     * Implementation of interface Runnable
     ****************************************************************************************/

    @Override
    public void run()
    {
        synchronized(mRobotExecutionMonitor)
        {
            if (!mRobotExecutionEnded)
            {
                try
                {
                    mRobotExecutionMonitor.wait();
                }
                catch (InterruptedException e) {}
            }
        }

        mRobotMonitoringThread = null;
        
        if (XSimulation.getMode() != ESimulationMode.HIDDEN)
        {
            if (ESessionState.RUNNING == mState) stop();
    
            if (ESessionState.TERMINATING != mState) msDelay(500);

            if (ESessionState.TERMINATING != mState) XAudioSystem.playRobotSignal();
    
            if (ESessionState.STOPPED == mState) complete();
            
            if (Env.ROBOT_IS_MASTER && (ESessionState.COMPLETED == mState))
            {
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                    	msDelay(1000);
                    	XSimulation.terminateSimulationProcess();
                    }
                });
            }
        }
    }
}
