/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.*;
import org.jabotics.pc.defs.SIMDefs;
import org.jabotics.robot.compile.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.util.Locale;

import javax.swing.SwingUtilities;

/**
 * Beschreiben Sie hier die Klasse XSimulation.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class XSimulation implements ISimulation
{
    static
    {
        // Initialization of constants defined in class Env depend on context.
        // => Set EnvInit.CONTEXT before accessing constants in class Env.
        EnvInit.CONTEXT = EContext.SIMULATION;
        // Initialize class Env if not yet done:
        Env.init();
    }

    // constants:
    static private final String ERR_MAIN = "org.jabotics.robot.sim.XSimulation: Invalid call to main method";
    static private final String USAGE = "Usage: java -jar jabotics-sim.jar [-l <language> [-c <country>]] [-d <robot root directory>] <robot class>";

    private static class OutErrStreamDrain implements Runnable
    {
    	private boolean mStop = false;

    	synchronized void stop()
    	{
    		mStop = true;
    		this.notifyAll();
    	}

    	public void run()
    	{
            PrintStream originalOut = System.out;
            PrintStream originalErr = System.err;

            InputStream isOut;
            InputStream isErr;

            try
            {
                PipedOutputStream osOut = new PipedOutputStream();
                isOut = new PipedInputStream(osOut);
                PrintStream psOut = new PrintStream(osOut);
                System.setOut(psOut);

                PipedOutputStream osErr = new PipedOutputStream();
                isErr = new PipedInputStream(osErr);
                PrintStream psErr = new PrintStream(osErr);
                System.setErr(psErr);
            }
            catch (IOException ioe)
            {
                System.setOut(originalOut);
                System.setErr(originalErr);
                return;
            }

    		try
    		{
    			synchronized(this)
    			{
		            while (!mStop)
		            {
		        		while (isOut.available()>0) isOut.read();
		        		while (isErr.available()>0) isErr.read();
		        		this.wait(10);
	        		}
    			}
            }
    		catch (IOException | InterruptedException e) {}

            System.setOut(originalOut);
            System.setErr(originalErr);
    	}
    }
    
    // class variables:
    static private Locale sLocale = null;
    static private String sRobotClassName = null;
    static private String sRobotRootPath = null;
    static private ESimulationMode sMode = ESimulationMode.NORMAL;
    static private XSimulation sSimulation = null;
    static private XSimulationServer sSimulationServer = null;
    static private volatile XSimulationView sSimulationView = null;
    static private volatile boolean sHasBeenTerminated = false;
    static private OutErrStreamDrain sOutErrStreamDrain = null;

    static public void main(String[] args) //throws Throwable
    {
        if (CS.LOG_BUFFER)
        {
            XLogBuffer.enableLogging(true);
        }

        String language = null;
        String country = null;

        if (Env.ROBOT_IS_MASTER)
        {
            String strMode = System.getProperty(SIMDefs.KEY_SIM_MODE);
            if (null == strMode)
            {
                throw new InternalError("Internal error in org.jabotics.robot.sim.XSimulation: Session mode undefined.");
            }
            try
            {
                int mode_id = Integer.parseInt(strMode);
                sMode = ESimulationMode.values()[mode_id];
            }
            catch (Exception e)
            {
                throw new IllegalStateException("Internal error in org.jabotics.robot.sim.XSimulation: Mode identifier not properly defined.");
            }

            if (CS.LOGGING_SIM) XLogBuffer.println("XSimulation.main(): Simulation mode set by robot master to: " + sMode.toString());
        }
        else
        {
            // For direct simulation invocation evaluate program arguments.
            // (sMode == ESimulationMode.NORMAL)

            int i = 0;
            while (i < args.length)
            {
                if (args[i].equals("-h"))
                {
                    if (args.length == 1)
                    {
                        printUsage();
                    }
                    else
                    {
                        printErrorMain();
                    }
                    return;
                }
                else if (args[i].equals("-l"))
                {
                    if (i == args.length)
                    {
                        printErrorMain();
                        return;
                    }
    
                    language = args[++i];
                    if (language.startsWith("-"))
                    {
                        printErrorMain();
                        return;
                    }
                }
                else if (args[i].equals("-c"))
                {
                    if (i == args.length)
                    {
                        printErrorMain();
                        return;
                    }
    
                    country = args[++i];
                    if (country.startsWith("-"))
                    {
                        printErrorMain();
                        return;
                    }
                }
                else if (args[i].equals("-d"))
                {
                    if (i == args.length)
                    {
                        printErrorMain();
                        return;
                    }
    
                    sRobotRootPath = args[++i];
                    if (sRobotRootPath.startsWith("-"))
                    {
                        printErrorMain();
                        return;
                    }
                }
                else
                {
                    sRobotClassName = args[i];
                    if ( (args.length - 1 != i) || sRobotClassName.startsWith("-") )
                    {
                        printErrorMain();
                        return;
                    }
                }
                ++i;
            }
    
            if ( (null == sRobotClassName) ||  (null == sRobotRootPath) || ( (null == language) && (null != country) ) )
            {
                printErrorMain();
                return;
            }

            if (CS.LOGGING_SIM)
            {
                XLogBuffer.println("XSimulation.main(): Simulation is master.");
                XLogBuffer.println("  sRobotClassName: " + sRobotClassName);
                XLogBuffer.println("  sRobotRootPath:  " + sRobotRootPath);
            }
        }

        sLocale = (null == language) ? null : ((null == country) ? new Locale(language) : new Locale(language,country));

        if (!Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
        {
            // Robot will run in this process and SIMAdapter expects session id as system property
            System.setProperty(SIMDefs.KEY_SIM_SESSION_ID, "0");
        }

        // Start simulation:
        XSimulation simulation = getSimulation();

        if (Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
        {
            try
            {
                if (CS.LOGGING_SIM) XLogBuffer.println("XSimulation.main(): Simulation server will be started.");

                sSimulationServer = new XSimulationServer(simulation);
            }
            catch (Exception e)
            {
                e.printStackTrace();
                System.exit(1);
            }
        }
    }

    static private void printErrorMain()
    {
        System.err.println(ERR_MAIN);
        System.err.println(USAGE);
    }

    static private void printUsage()
    {
        System.out.println(USAGE);
    }

    static Locale getLocale()
    {
        return sLocale;
    }

    // This method will be called to stop the simulation process.
    static void terminateSimulationProcess()
    {
        shutdown(); // will cause the UI to be disposed and robot threads or processes to stop

        System.exit(0); // Seems to be necessary to stop the AWT event thread.
    }

    static ESimulationMode getMode()
    {
        return sMode;
    }

    static synchronized ESimulationMode setMode(ESimulationMode mode)
    {
        if (null == sSimulation)
        {
            sMode = mode;
        }
        return mode;
    }

    static void startOutErrStreamDrain()
    {
    	if (null == sOutErrStreamDrain)
    	{
			sOutErrStreamDrain = new OutErrStreamDrain();
			Thread thread = new Thread(sOutErrStreamDrain);
			thread.setDaemon(true);
			thread.start();
    	}
    }

    static void stopOutErrStreamDrain()
    {
    	if (null != sOutErrStreamDrain)
    	{
			sOutErrStreamDrain.stop();
			sOutErrStreamDrain = null;
    	}
    }


    static synchronized XSimulation getSimulation()
    {
        if (null == sSimulation)
        {
            if (CS.LOG_BUFFER)
            {
                XLogBuffer.enableLogging(true);
            }

            sSimulation = new XSimulation();
        }

        return sSimulation;
    }


    final boolean mInteractive;
    XSession mSession = XSimulationModel.getNewRobotSession(sRobotClassName,sRobotRootPath);
    XSensorRef.SensorProxy mSensorProxy = XSensorRef.getSensorProxy();
    XDisplay mDisplay = mSession.getRobot().getDisplay();
    XSpeaker mSpeaker = mSession.getRobot().getSpeaker();
    XConsole mConsole = XConsole.getConsole();
    XRobot mRobot = mSession.getRobot();

    private XSimulation()
    {
        if (null == sLocale) sLocale = Locale.getDefault();
        XMsg.init(sLocale);

        XPreferences.enableTestMode(ESimulationMode.NORMAL != sMode);
        XAudioSystem.enableSilentMode(ESimulationMode.HIDDEN == sMode);

        mInteractive = ((ESimulationMode.NORMAL == sMode) || (ESimulationMode.TEST == sMode));

        // In interactive mode, the session has to be initialized once before displaying the UI.
        // Further session handling will be controlled by the UI.
        if (mInteractive) mSession.init();

        if (ESimulationMode.HIDDEN != sMode)
        {
            // TODO:
    	    // Currently, BlueJ displays JavaFX related error messages
    	    // when starting the simulation by invoking the main method
    	    // of the robot class for debugging purposes. For the moment,
    		// remove everything from the streams until UI is up and running.
        	if (ESimulationMode.NORMAL != sMode)  startOutErrStreamDrain();

        	SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    sSimulationView = new XSimulationView("Jabotics - " + XMsg.ROBOT_SIM + " - " + sRobotClassName, mInteractive);
                    sSimulationView.setVisible(true);
                }
            });
        }
    }

    static public void shutdown()
    {
        if (!sHasBeenTerminated)
        {
            sHasBeenTerminated = true;

            // Simulation view should already be disposed if simulation has
            // been closed via the UI. For other cases:
            // Running simulation view causes trouble during shutdown of VM
            // (see documentation of AWT threading issues), therefore:
            if (null != sSimulationView)
            {
                sSimulationView.dispose();
                sSimulationView = null;
            }
    
            XSession[] sessions = XSimulationModel.getRobotSessions();
            for (XSession session : sessions)
            {
                // If robot program is running in separate process controlled by simulation tool,
                // session.terminate() will wait for the end of the forked robot process. 
                session.terminate();
            }
    
            if (null != sSimulationServer)
            {
                sSimulationServer.terminate();
                sSimulationServer = null;
            }
        }
    }

    /*********************************************************************************************
     * methods of ISessionControl
     *********************************************************************************************/

    @Override
    public String simulationGetUUID()
    {
        return Env.SIM_UUID;
    }

    @Override
    public boolean sessionSuspendRobotThreadIsEnabled(int session_id)
    {
        return (((ESimulationMode.NORMAL == sMode) || (ESimulationMode.TEST == sMode)) && XPreferences.getReadInstance().getSimSuspendEnabled());
    }

    @Override
    public void sessionResetPreferences(int session_id)
    {
        XSimulationModel.getRobotSession(session_id).resetPreferences();
    }

    @Override
    public void sessionImportPreferences(int session_id, String pathname)
    {
        XSimulationModel.getRobotSession(session_id).importPreferences(pathname);
    }

    @Override
    public void sessionImportPreferencesByClass(int session_id, String classname)
    {
        XSimulationModel.getRobotSession(session_id).importPreferencesByClass(classname);
    }

    @Override
    public void sessionInit(int session_id)
    {
        XSimulationModel.getRobotSession(session_id).init();
    }

    @Override
    public void sessionStart(int session_id)
    {
        XSimulationModel.getRobotSession(session_id).start();
    }

    @Override
    public void sessionStop(int session_id)
    {
        XSimulationModel.getRobotSession(session_id).stop();
    }

    @Override
    public void sessionComplete(int session_id)
    {
        XSimulationModel.getRobotSession(session_id).complete();
    }

    @Override
    public void sessionTerminate(int session_id)
    {
        XSimulationModel.getRobotSession(session_id).terminate();
    }

    @Override
    public void sessionRobotExecutionEnded(int session_id)
    {
        if (CS.LOGGING_SIM) XLogBuffer.println("XSimulation.sessionRobotExecutionEnded(): Called for session " + session_id);
        XSimulationModel.getRobotSession(session_id).robotExecutionEnded();
    }

    @Override
    public void sessionPressButton(int session_id, int button_id, int period)
    {
        XSimulationModel.getRobotSession(session_id).pressButton(EButtonType.values()[button_id], period);
    }

    @Override
    public void sessionClap(int volume)
    {
        XSimulationModel.getSoundscape().addClap(volume);
    }

    @Override
    public void sessionSetNoiseLevel(int level)
    {
        XSimulationModel.getSoundscape().setNoiseLevel(level);
    }

    /*********************************************************************************************
     * methods of ISimulation
     *********************************************************************************************/

    @Override
    public long[] sessionGetData(int session_id)
    {
        return XSimulationModel.getRobotSession(session_id).getSessionData().getData();
    }

    @Override
    public XSensorInfo sensorRegisterSensor(int session_id, int port_id, int type_id)
    {
        return mSensorProxy.registerSensor(port_id, type_id);
    }

    @Override
    public XSensorInfo sensorRegisterButton(int session_id, int button_id)
    {
        return mSensorProxy.registerButton(button_id);
    }

    @Override
    public void sensorSwitchLightOn(int session_id, int port_id)
    {
        mSensorProxy.switchLightOn(port_id);
    }

    @Override
    public void sensorSwitchLightOff(int session_id, int port_id)
    {
        mSensorProxy.switchLightOff(port_id);
    }

    @Override
    public float[] sensorGetValues(int session_id)
    {
        return mSensorProxy.getMeasuredValues();
    }

    @Override
    public void displayRefresh(int session_id)
    {
        mDisplay.refresh();
    }

    @Override
    public void displayClear(int session_id)
    {
        mDisplay.clear();
    }

    @Override
    public void displayClear(int session_id, int y_position)
    {
        mDisplay.clear(y_position);
    }

    @Override
    public void displayClear(int session_id, int x_position, int y_position, int count)
    {
        mDisplay.clear(x_position, y_position, count);
    }

    @Override
    public void displayDraw(int session_id, char ch, int x_position, int y_position)
    {
        mDisplay.draw(ch, x_position, y_position);
    }

    @Override
    public void displayDraw(int session_id, int n, int x_position, int y_position)
    {
        mDisplay.draw(n, x_position, y_position);
    }

    @Override
    public void displayDraw(int session_id, int n, int x_position, int y_position, int digits)
    {
        mDisplay.draw(n, x_position, y_position, digits);
    }

    @Override
    public void displayDraw(int session_id, float fl, int x_position, int y_position)
    {
        mDisplay.draw(fl, x_position, y_position);
    }

    @Override
    public void displayDraw(int session_id, String text, int x_position, int y_position)
    {
        mDisplay.draw(text, x_position, y_position);
    }

    @Override
    public void displayDraw(int session_id, String text, int x_position, int y_position, boolean inverted)
    {
        mDisplay.draw(text, x_position, y_position, inverted);
    }

    @Override
    public int soundGetVolume(int session_id)
    {
        return mSpeaker.getVolume();
    }

    @Override
    public void soundSetVolume(int session_id, int volume)
    {
        mSpeaker.setVolume(volume);
    }

    @Override
    public int soundBeep(int session_id)
    {
        return mSpeaker.beep();
    }

    @Override
    public int soundTwoBeeps(int session_id)
    {
    	return mSpeaker.twoBeeps();
    }

    @Override
    public int soundBeepSequence(int session_id)
    {
    	return mSpeaker.beepSequence();
    }

    @Override
    public int soundBeepSequenceUp(int session_id)
    {
    	return mSpeaker.beepSequenceUp();
    }

    @Override
    public int soundBuzz(int session_id)
    {
    	return mSpeaker.buzz();
    }

    @Override
    public void soundTone(int session_id, int frequency, int period)
    {
        mSpeaker.tone(frequency, period);
    }

    @Override
    public void soundTone(int session_id, int frequency, int period, int volume)
    {
        mSpeaker.tone(frequency, period, volume);
    }

    @Override
    public void soundNote(int session_id, int instrument_id, int frequency, int period)
    {
        mSpeaker.note(ESound.values()[instrument_id], frequency, period);
    }

    @Override
    public void consoleOpen(int session_id)
    {
        mConsole.open();
    }

    @Override
    public void consoleClose(int session_id)
    {
        mConsole.close();
    }

    @Override
    public void consolePrint(int session_id, String text)
    {
        mConsole.print(text);
    }

    @Override
    public void consolePrintln(int session_id, String text)
    {
        mConsole.println(text);
    }

    @Override
    public float motorGetMaxSpeed(int session_id, int port_id)
    {
        return mRobot.getMotorSet().getMotor(EMotorPort.values()[port_id]).getMaxSpeed();
    }

    @Override
    public void motorStartControl(int session_id, int port_id)
    {
        mRobot.getMotorSet().getMotor(EMotorPort.values()[port_id]).startControl();
    }

    @Override
    public void motorSetToFloatMode(int session_id, int port_id)
    {
        mRobot.getMotorSet().getMotor(EMotorPort.values()[port_id]).setToFloatMode();
    }

    @Override
    public float[] motorGetPIDParameters(int session_id, int port_id)
    {
        return mRobot.getMotorSet().getMotor(EMotorPort.values()[port_id]).getPIDParameters();
    }

    @Override
    public float[] motorGetTachoCounts(int session_id)
    {
        return mRobot.getMotorSet().getTachoCounts();
    }

    @Override
    public void motorSetAllPowerData(int session_id, float[] power)
    {
        for (EMotorPort port : EMotorPort.values())
        {
            mRobot.getMotorSet().getMotor(port).setPower(power[port.ordinal()]);
        }
    }

    @Override
    public float robotGetDefaultTrackWidth(int session_id)
    {
        return mRobot.getDefaultTrackWidth();
    }

    @Override
    public float robotGetDefaultWheelDiameter(int session_id)
    {
        return mRobot.getDefaultWheelDiameter();
    }

    @Override
    public void robotSetTrackWidth(int session_id, float track_width)
    {
        mRobot.setTrackWidth(track_width);
    }

    @Override
    public void robotSetWheelDiameter(int session_id, float wheel_diameter)
    {
        mRobot.setWheelDiameter(wheel_diameter);
    }

    @Override
    public void robotSetMotors(int session_id, int left_motor_port_id, int right_motor_port_id)
    {
        mRobot.setMotors(EMotorPort.values()[left_motor_port_id], EMotorPort.values()[right_motor_port_id]);
    }
}
