/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.util.XTimedThread;

import java.util.LinkedList;
import java.util.Vector;

/**
 * Beschreiben Sie hier die Klasse XSimulationModel.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class XSimulationModel
{
	// Constants:
    static private final int DYNAMICS_UPDATE_DELAY = 1;


    public interface IPostUpdateHook
    {
        public void postDynamicsUpdate();
    }

    
    static class XDynamicsUpdater implements XTimedThread.IRunnable
    {
        private XDynamicsUpdater()
        {
        }

        @Override
        public void init()
        {
        }

        @Override
        public void runTimed()
        {
            LinkedList<BModelElement> elements = sUpdateElements;
            for (BModelElement element : elements)
            {
                element.updateDynamics();
            }

            LinkedList<IPostUpdateHook> hooks = sPostUpdateHooks;
            for (IPostUpdateHook hook : hooks)
            {
                hook.postDynamicsUpdate();
            }
        }

        @Override
        public void terminate()
        {
            // nothing to do since runTimed() should finish quickly
        }
    
        @Override
        public void shutdown()
        {
            //DEBUG AConsole.close();

            // Fenster schließen?
        }

    }

    static synchronized void addUpdateElement(BModelElement element)
    {
    	LinkedList<BModelElement> elements = new LinkedList<BModelElement>(sUpdateElements);
        if (!elements.contains(element)) elements.add(element);
        sUpdateElements = elements;
    }

    static synchronized void removeUpdateElement(BModelElement element)
    {
    	LinkedList<BModelElement> elements = new LinkedList<BModelElement>(sUpdateElements);
        elements.remove(element);
        sUpdateElements = elements;
    }

    static synchronized void addPostUpdateHook(IPostUpdateHook hook)
    {
    	LinkedList<IPostUpdateHook> hooks = new LinkedList<IPostUpdateHook>(sPostUpdateHooks);
        if (!hooks.contains(hook)) hooks.add(hook);
        sPostUpdateHooks = hooks;
    }

    static synchronized void removePostUpdateHook(IPostUpdateHook hook)
    {
    	LinkedList<IPostUpdateHook> hooks = new LinkedList<IPostUpdateHook>(sPostUpdateHooks);
        hooks.remove(hook);
        sPostUpdateHooks = hooks;
    }


    static synchronized XSession getNewRobotSession(String robotClassName, String robotRootPath)
    {
		XSession session;

		Vector<XSession> robotSessions = new Vector<XSession>(sRobotSessions);
		// find first free index:
		int i = 0;
		for (; i < robotSessions.size(); ++i) if (null == robotSessions.elementAt(i)) break;
		session = new XSession(i, robotClassName, robotRootPath);
		if (i < robotSessions.size())
		{
			robotSessions.setElementAt(session, i);
		}
		else
		{
			robotSessions.addElement(session);
		}
    	session.addSessionHandler(sSoundscape.getSessionHandler());
		sRobotSessions = robotSessions;
    	
		return session;
    }

    static XSession[] getRobotSessions()
    {
    	Vector<XSession> robotSessions = sRobotSessions;
		int size = 0;
		for (int i = 0; i < robotSessions.size(); ++i) if (null != robotSessions.elementAt(i)) ++size;
		XSession[] sessions = new XSession[size];
		for (int i = 0, j = 0; i < robotSessions.size(); ++i)
		{
			XSession session = robotSessions.elementAt(i);
			if (null != session) sessions[j++] = session;
		}
		return sessions;
    }

    static XSession getRobotSession(int session_id)
    {
    	return sRobotSessions.elementAt(session_id);
    }

    static synchronized void removeRobotSession(XSession session)
    {
    	session.removeSessionHandler(sSoundscape.getSessionHandler().getKey());
    	Vector<XSession> robotSessions = new Vector<XSession>(sRobotSessions);
    	robotSessions.remove(session.getID());
    	sRobotSessions = robotSessions;
    }

    
    // Klassenvariablen
    static private volatile Vector<XSession> sRobotSessions = new Vector<XSession>(10);
    static private volatile LinkedList<BModelElement> sUpdateElements = new LinkedList<BModelElement>();
    static private volatile LinkedList<IPostUpdateHook> sPostUpdateHooks = new LinkedList<IPostUpdateHook>();

    static private final XDynamicsUpdater sDynamicsUpdater = new XDynamicsUpdater();
    static private final XTimedThread sDynamicsUpdateThread = new XTimedThread("Dynamics Update Thread",sDynamicsUpdater);
    static private final XScene sScene = new XScene();
    static private final XSoundscape sSoundscape = new XSoundscape();
    static
    {
        // Vorbereiten und Starten des Steuerungs-Threads:
        sDynamicsUpdateThread.enableTimeDilatation(false);
        sDynamicsUpdateThread.setPriority(Thread.MAX_PRIORITY - 1);
        sDynamicsUpdateThread.setDaemon(true);
        sDynamicsUpdateThread.setTimeouts(DYNAMICS_UPDATE_DELAY,DYNAMICS_UPDATE_DELAY,0);
        sDynamicsUpdateThread.start();
    }

    static public XDynamicsUpdater getDynamicsUpdater()
    {
        return sDynamicsUpdater;
    }
    
    static public long getDynamicsUpdateCount()
    {
    	return sDynamicsUpdateThread.getMonitor().getCalls();
    }

    static void stopDynamicsUpdateThread()
    {
        sDynamicsUpdateThread.terminate();
    }

    static public XScene getScene()
    {
        return sScene;
    }

    static public XSoundscape getSoundscape()
    {
    	return sSoundscape;
    }

    private XSimulationModel() {}

    
    /*static public void main(String[] args)
    {
    	XAudioSystem.testSDL();
    	XSpeaker.testTone();
    }*/
}
