/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NoSuchObjectException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;

import org.jabotics.robot.compile.CS;
import org.jabotics.robot.util.XLogBuffer;

/**
 * @author  Frank Gollnik
 * @version 1.2.1
 */
public class XSimulationServer extends UnicastRemoteObject implements ISimulationServer
{
    private static final long serialVersionUID = 8277892480130009004L;

    static
    {
        if (Env.ROBOT_RUNS_IN_SEPARATE_PROCESS)
        {
            try
            {
                // Make sure that rmi registry is running
                LocateRegistry.createRegistry(Env.RMI_PORT);
            }
            catch (Exception e) {}
        }
    }


    private final XSimulation mSimulation;
    

    XSimulationServer(XSimulation simulation) throws RemoteException
    {
        super(Env.RMI_PORT);

        mSimulation = simulation;

        try
        {
            Naming.rebind(Env.RMI_SERVER, this);
        }
        catch (MalformedURLException e)
        {
            throw new InternalError("Internal error in XSimulationServer: Malformed URL " + Env.RMI_SERVER);
        }

        if (CS.LOGGING_SIM) XLogBuffer.println("XSimulationServer(): Started with port " + Env.RMI_PORT);
    }

    void unbind()
    {
        try
        {
            Naming.unbind(Env.RMI_SERVER);
        }
        catch (RemoteException | MalformedURLException | NotBoundException e)
        {
            e.printStackTrace();
            System.exit(1);
        }
    }

    void terminate()
    {
        try
        {
            UnicastRemoteObject.unexportObject(this, true);
        }
        catch (NoSuchObjectException e) {}
    }


    /*********************************************************************************************
     * Implementation of ISimulationServer
     *********************************************************************************************/

    @Override
    public String simulationGetUUID() throws RemoteException
    {
        return mSimulation.simulationGetUUID();
    }

    @Override
    public boolean sessionSuspendRobotThreadIsEnabled(int session_id) throws RemoteException
    {
        return mSimulation.sessionSuspendRobotThreadIsEnabled(session_id);
    }

    @Override
    public long[] sessionGetData(int session_id) throws RemoteException
    {
        return mSimulation.sessionGetData(session_id);
    }

    @Override
    public void sessionResetPreferences(int session_id) throws RemoteException
    {
        mSimulation.sessionResetPreferences(session_id);
    }

    @Override
    public void sessionImportPreferences(int session_id, String pathname) throws RemoteException
    {
        mSimulation.sessionImportPreferences(session_id, pathname);
    }

    @Override
    public void sessionImportPreferencesByClass(int session_id, String classname) throws RemoteException
    {
        mSimulation.sessionImportPreferencesByClass(session_id, classname);
    }

    @Override
    public void sessionInit(int session_id) throws RemoteException
    {
        mSimulation.sessionInit(session_id);
    }

    @Override
    public void sessionStart(int session_id) throws RemoteException
    {
        mSimulation.sessionStart(session_id);
    }

    @Override
    public void sessionStop(int session_id) throws RemoteException
    {
        mSimulation.sessionStop(session_id);
    }

    @Override
    public void sessionComplete(int session_id) throws RemoteException
    {
        mSimulation.sessionComplete(session_id);
    }

    @Override
    public void sessionTerminate(int session_id) throws RemoteException
    {
        mSimulation.sessionTerminate(session_id);
    }

    @Override
    public void sessionRobotExecutionEnded(int session_id) throws RemoteException
    {
        mSimulation.sessionRobotExecutionEnded(session_id);
    }

    @Override
    public void sessionPressButton(int session_id, int button_id, int period) throws RemoteException
    {
        mSimulation.sessionPressButton(session_id, button_id, period);
    }

    @Override
    public void sessionClap(int volume) throws RemoteException
    {
        mSimulation.sessionClap(volume);
    }

    @Override
    public void sessionSetNoiseLevel(int volume) throws RemoteException
    {
        mSimulation.sessionSetNoiseLevel(volume);
    }

    @Override
    public XSensorInfo sensorRegisterSensor(int session_id, int port_id, int type_id) throws RemoteException
    {
        return mSimulation.sensorRegisterSensor(session_id, port_id, type_id);
    }

    @Override
    public XSensorInfo sensorRegisterButton(int session_id, int button_id) throws RemoteException
    {
        return mSimulation.sensorRegisterButton(session_id, button_id);
    }

    @Override
    public void sensorSwitchLightOn(int session_id, int port_id) throws RemoteException
    {
        mSimulation.sensorSwitchLightOn(session_id, port_id);
    }

    @Override
    public void sensorSwitchLightOff(int session_id, int port_id) throws RemoteException
    {
        mSimulation.sensorSwitchLightOff(session_id, port_id);
    }

    @Override
    public float[] sensorGetValues(int session_id) throws RemoteException
    {
        return mSimulation.sensorGetValues(session_id);
    }

    @Override
    public void displayRefresh(int session_id) throws RemoteException
    {
        mSimulation.displayRefresh(session_id);
    }

    @Override
    public void displayClear(int session_id) throws RemoteException
    {
        mSimulation.displayClear(session_id);
    }

    @Override
    public void displayClear(int session_id, int y_position) throws RemoteException
    {
        mSimulation.displayClear(session_id, y_position);
    }

    @Override
    public void displayClear(int session_id, int x_position, int y_position, int count) throws RemoteException
    {
        mSimulation.displayClear(session_id, x_position, y_position, count);
    }

    @Override
    public void displayDraw(int session_id, char ch, int x_position, int y_position) throws RemoteException
    {
        mSimulation.displayDraw(session_id, ch, x_position, y_position);
    }

    @Override
    public void displayDraw(int session_id, int n, int x_position, int y_position) throws RemoteException
    {
        mSimulation.displayDraw(session_id, n, x_position, y_position);
    }

    @Override
    public void displayDraw(int session_id, int n, int x_position, int y_position, int digits) throws RemoteException
    {
        mSimulation.displayDraw(session_id, n, x_position, y_position, digits);
    }

    @Override
    public void displayDraw(int session_id, float fl, int x_position, int y_position) throws RemoteException
    {
        mSimulation.displayDraw(session_id, fl, x_position, y_position);
    }

    @Override
    public void displayDraw(int session_id, String text, int x_position, int y_position) throws RemoteException
    {
        mSimulation.displayDraw(session_id, text, x_position, y_position);
    }

    @Override
    public void displayDraw(int session_id, String text, int x_position, int y_position, boolean inverted) throws RemoteException
    {
        mSimulation.displayDraw(session_id, text, x_position, y_position, inverted);
    }

    @Override
    public int soundGetVolume(int session_id) throws RemoteException
    {
        return mSimulation.soundGetVolume(session_id);
    }

    @Override
    public void soundSetVolume(int session_id, int volume) throws RemoteException
    {
        mSimulation.soundSetVolume(session_id, volume);
    }

    @Override
    public int soundBeep(int session_id) throws RemoteException
    {
        return mSimulation.soundBeep(session_id);
    }

    @Override
    public int soundTwoBeeps(int session_id) throws RemoteException
    {
    	return mSimulation.soundTwoBeeps(session_id);
    }

    @Override
    public int soundBeepSequence(int session_id) throws RemoteException
    {
    	return mSimulation.soundBeepSequence(session_id);
    }

    @Override
    public int soundBeepSequenceUp(int session_id) throws RemoteException
    {
    	return mSimulation.soundBeepSequenceUp(session_id);
    }

    @Override
    public int soundBuzz(int session_id) throws RemoteException
    {
    	return mSimulation.soundBuzz(session_id);
    }

    @Override
    public void soundTone(int session_id, int frequency, int period) throws RemoteException
    {
        mSimulation.soundTone(session_id, frequency, period);
    }

    @Override
    public void soundTone(int session_id, int frequency, int period, int volume) throws RemoteException
    {
        mSimulation.soundTone(session_id, frequency, period, volume);
    }

    @Override
    public void soundNote(int session_id, int instrument_id, int frequency, int period) throws RemoteException
    {
        mSimulation.soundNote(session_id, instrument_id, frequency, period);
    }

    @Override
    public void consoleOpen(int session_id) throws RemoteException
    {
        mSimulation.consoleOpen(session_id);
    }

    @Override
    public void consoleClose(int session_id) throws RemoteException
    {
        mSimulation.consoleClose(session_id);
    }

    @Override
    public void consolePrint(int session_id, String text) throws RemoteException
    {
        mSimulation.consolePrint(session_id, text);
    }

    @Override
    public void consolePrintln(int session_id, String text) throws RemoteException
    {
        mSimulation.consolePrintln(session_id, text);
    }

    @Override
    public float motorGetMaxSpeed(int session_id, int port_id) throws RemoteException
    {
        return mSimulation.motorGetMaxSpeed(session_id, port_id);
    }

    @Override
    public void motorStartControl(int session_id, int port_id) throws RemoteException
    {
        mSimulation.motorStartControl(session_id, port_id);
    }

    @Override
    public void motorSetToFloatMode(int session_id, int port_id) throws RemoteException
    {
        mSimulation.motorSetToFloatMode(session_id, port_id);
    }

    @Override
    public float[] motorGetPIDParameters(int session_id, int port_id) throws RemoteException
    {
        return mSimulation.motorGetPIDParameters(session_id, port_id);
    }

    @Override
    public float[] motorGetTachoCounts(int session_id) throws RemoteException
    {
        return mSimulation.motorGetTachoCounts(session_id);
    }

    @Override
    public void motorSetAllPowerData(int session_id, float[] power) throws RemoteException
    {
        mSimulation.motorSetAllPowerData(session_id, power);
    }

    @Override
    public float robotGetDefaultTrackWidth(int session_id) throws RemoteException
    {
        return mSimulation.robotGetDefaultTrackWidth(session_id);
    }

    @Override
    public float robotGetDefaultWheelDiameter(int session_id) throws RemoteException
    {
        return mSimulation.robotGetDefaultWheelDiameter(session_id);
    }

    @Override
    public void robotSetTrackWidth(int session_id, float track_width) throws RemoteException
    {
        mSimulation.robotSetTrackWidth(session_id, track_width);
    }

    @Override
    public void robotSetWheelDiameter(int session_id, float wheel_diameter) throws RemoteException
    {
        mSimulation.robotSetWheelDiameter(session_id, wheel_diameter);
    }

    @Override
    public void robotSetMotors(int session_id, int left_motor_port_id, int right_motor_port_id) throws RemoteException
    {
        mSimulation.robotSetMotors(session_id, left_motor_port_id, right_motor_port_id);
    }
}
