/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.util.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
//import javax.sound.sampled.*;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Beschreiben Sie hier die Klasse XSoundscape.
 * 
 * @author  Frank Gollnik
 * @version 1.2.1
 */
class XSoundscape implements XTimedThread.IRunnable, XPreferences.IListener
{
    // constants and class variables:
    static private final int MIXER_THREAD_PRIORITY = Thread.MAX_PRIORITY - 3;
    static private final int UPDATE_DELAY = 1;
    static private final float NOISE_AMPLITUDE = 10f;
    static private final float NOISE_PERIOD = 0.05f; // in s
    static private final float CLAP_PERIOD = 0.15f;  // in s

    private class SessionHandler extends BSessionHandler
    {
    	@Override
        public String getKey()
        {
            return "XSoundscape";
        }

        @Override
        public void startSession()
        {
        	if (0 == mRunningSessionsCnt.getAndIncrement())
        	{
        		mSysStartTime = 1e-9 * System.nanoTime() - mTPaused;
        		mThread.resume();
        	}
        }

        @Override
        public void pauseSession()
        {
        	if (0 == mRunningSessionsCnt.decrementAndGet())
        	{
        		mThread.pause();
        		mTPaused = 1e-9 * System.nanoTime() - mSysStartTime;
        	}
        }

        @Override
        public void resumeSession()
        {
        	if (0 == mRunningSessionsCnt.getAndIncrement())
        	{
        		mSysStartTime = 1e-9 * System.nanoTime() - mTPaused;
        		mThread.resume();
        	}
        }

        @Override
        public void stopSession()
        {
        	if (0 == mRunningSessionsCnt.decrementAndGet())
        	{
        		mThread.pause();
        		mTPaused = 1e-9 * System.nanoTime() - mSysStartTime;
        	}
        }
    }

    private class SoundInfo
    {
        ArrayList<Float> mLevels = new ArrayList<Float>();
        int mTotalLevel = -1;
    }

    private class SoundSource implements TICloneable<SoundSource>
    {
        private float mLevel;         // start/average sound level
        private float mAmplitude;     // modulation depth of the sound level
        private double mPeriod;       // in s
        private double mTargetTime;   // in s
        private float mTargetLevel;
        private float mCurrentLevel;

        @Override
        public void copyDataFrom(SoundSource template)
        {
            if (this != template)
            {
                mLevel = template.mLevel;
                mAmplitude = template.mAmplitude;
                mPeriod = template.mPeriod;
                mTargetTime = template.mTargetTime;
                mTargetLevel = template.mTargetLevel;
                mCurrentLevel = template.mCurrentLevel;
            }
        }
        
        @Override
        public SoundSource getClone()
        {
    	    SoundSource clone = new SoundSource();
    	    clone.copyDataFrom(this);
            return clone;
        }

        void generateClap(float level, double period)
        {
            mLevel = 0f;
            mAmplitude = 0f;
            mPeriod = period;
            mTargetTime = 0.;
            mTargetLevel = level;
            mCurrentLevel = 0f;
        }

        void generateNoise(float level, float amplitude, double period)
        {
            mLevel = level;
            mAmplitude = amplitude;
            mPeriod = period;
            mTargetTime = 0.;
            mTargetLevel = level;
            mCurrentLevel = level;
        }

        boolean update()
        {
        	if (mTargetTime < mT)
            {
                if (mLevel <= 0)
                {
                    // clap

                	// second phase is over ? => done
                	if (mTargetLevel <= 0) return true;

                    // start second phase (decay)
                    mTargetLevel = 0f;
                    mTargetTime = mT + 10. * mPeriod;
                }
                else
                {
                    // noise
                    mTargetLevel = mLevel + (float)(2.*Math.random() - 1.) * mAmplitude;
                    mTargetTime = mT + mPeriod;
                }
            }
            else
            {
                float factorReplacing = (float)(mDT / mPeriod);
                if (factorReplacing > 1f) factorReplacing = 1f;
                float factorRemaining = 1f - factorReplacing;
                mCurrentLevel = factorRemaining * mCurrentLevel + factorReplacing * mTargetLevel;
            }
            return false;
        }
    }

    private final SessionHandler mSessionHandler;
    private final XTimedThread mThread;
    private volatile LinkedList<SoundSource> mSources;
    private volatile SoundInfo mSoundInfo;
    private volatile int mInputLevel;
    private volatile int mMicSignal;
    private volatile TPipe<SoundSource> mSoundPipe = new TPipe<SoundSource>(new SoundSource());
    private volatile double mSysStartTime;
    private volatile double mT;
    private volatile double mTLast;
    private volatile double mDT;
    private volatile double mTPaused;
	private AtomicInteger mRunningSessionsCnt = new AtomicInteger(0);

    XSoundscape()
    {
        initVolatileMembers();
        mSessionHandler = new SessionHandler();
        mThread = new XTimedThread("XSoundscape controler thread",this);
        mThread.enableTimeDilatation(false);
        mThread.setDaemon(true);
        mThread.setPriority(MIXER_THREAD_PRIORITY);
        mThread.setTimeouts(UPDATE_DELAY,UPDATE_DELAY,0);
        mThread.pause(); // initially pausing
        mThread.start();
        XPreferences.addListener(this);
    }
    
    private void initVolatileMembers()
    {
        XPreferences.Read p = XPreferences.getReadInstance();
        mSources = new LinkedList<SoundSource>();
        mSoundInfo = new SoundInfo();
        mInputLevel = p.getAudioInputLevel();
        mMicSignal = 0;
        mSoundPipe = new TPipe<SoundSource>(new SoundSource());
        mSysStartTime = 1e-9 * System.nanoTime();
        mT = 0.0;
        mTLast = 0.0;
        mDT = 0.0;
        mTPaused = 0.0;
        setNoise(mInputLevel);
    }

    ISessionHandler getSessionHandler()
    {
    	return mSessionHandler;
    }

    void setNoiseLevel(int level)
    {
        if (level <= 0) return;
        if (level > 100) level = 100;
    	setNoise(level);
    }

    private void setNoise(float level)
    {
        synchronized (mSoundPipe)
        {
        	mSoundPipe.getWriteObject().generateNoise(level, (level < NOISE_AMPLITUDE) ? level : NOISE_AMPLITUDE, NOISE_PERIOD);
        	mSoundPipe.commitWriteObject();
        }
    	this.runTimed();
    }

    void addClap(int level)
    {
        if (level <= 0) return;
        if (level > 100) level = 100;
        addClap(1.6f*level,CLAP_PERIOD); // empirical factor to reach passed value as peak value ( 1/(1-1/e) = 1.58 )
    }

    private void addClap(float level, double period)
    {
        synchronized (mSoundPipe)
        {
        	mSoundPipe.getWriteObject().generateClap(level, period);
        	mSoundPipe.commitWriteObject();
        }
    	this.runTimed();
    }

    @Override
    public synchronized void init()
    {
        this.runTimed(); // single call to complete initialization
    }

    @Override
    public synchronized void shutdown()
    {
    }

    @Override
    public synchronized void terminate()
    {
    }

    @Override
    public void runTimed()
    {
    	LinkedList<SoundSource> sources = mSources;
    	synchronized(sources)
    	{
            mTLast = mT;
            mT = 1e-9 * System.nanoTime() - mSysStartTime;
            mDT = mT - mTLast;

            mMicSignal = XAudioSystem.getMicSignal();
	
	        // process list of new tasks in any case (preferences could change: keep track of sound input level respectively noise level)
	        TPipe<SoundSource> soundPipe = mSoundPipe;
	        while (soundPipe.hasNext())
	        {
	            SoundSource soundSource = soundPipe.read().getClone();
	            soundSource.mTargetTime = mT + soundSource.mPeriod;
	            if (soundSource.mLevel > 0)
	            {
	            	// noise => remove existing noise entry
		            for (Iterator<SoundSource> it = sources.iterator(); it.hasNext(); )
		            {
		                SoundSource src = it.next();
		                if (src.mLevel > 0)
		                {
		                    // completed
		                    it.remove();
		                }
		            }
	            }
	            sources.add(soundSource);
	        }

	        if (mMicSignal < 0)
	        {
	            // mic is off  =>  simulate noise / clap
	            SoundInfo info = new SoundInfo();
	            for (Iterator<SoundSource> it = sources.iterator(); it.hasNext(); )
	            {
	                SoundSource src = it.next();
	                if (src.update())
	                {
	                    // completed
	                    it.remove();
	                }
	                else
	                {
	                    info.mLevels.add(src.mCurrentLevel);
	                }
	            }
	            mSoundInfo = info;
	        }
    	}
    }

    int getSoundLevel()
    {
        int micSignal = mMicSignal;

        if (micSignal < 0)
        {
            SoundInfo info = mSoundInfo;
            synchronized(info)
            {
                if (info.mTotalLevel < 0)
                {
                    double totalLevel = 0.;
                    for (float level : info.mLevels)
                    {
                        if (level > 0) totalLevel += Math.exp(level);
                    }
                    info.mTotalLevel = (totalLevel > 0) ? (int)Math.log(totalLevel) : 0;
                    if (info.mTotalLevel < 0)
                    {
                        info.mTotalLevel = 0;
                    }
                    else if (info.mTotalLevel > 100)
                    {
                        info.mTotalLevel = 100;
                    }
                }

                return info.mTotalLevel;
            }
        }
        else
        {
            int level = (micSignal * mInputLevel) / 50;  // => max "amplification" by factor of 100/50 = 2
            return (level < 0) ? 0 : ( (level > 100) ? 100 : level );
        }
    }

    public void preferencesChanged(XPreferences.Read newPrefs)
    {
        if (mInputLevel != newPrefs.getAudioInputLevel())
        {
            mInputLevel = newPrefs.getAudioInputLevel();
            this.setNoise(mInputLevel);
            /*if (ESessionState.RUNNING != mSession.getState())
            {
                runTimed(); // manual call necessary in this case to complete initialization
            }*/
        }
    }

    /*static public void test1()
    {
        XSoundscape soundscape = ???.getSoundscape();
        soundscape.runTimed();
        int level = soundscape.getSoundLevel();
        System.out.println("Sound level = " + level);
    }*/
}
